from inspect import signature
from collections import namedtuple

import jax
import jax.numpy as jnp
import numpy as onp
import haiku as hk
from gym.spaces import Space, Discrete

from ..utils import safe_sample, batch_to_single, default_preprocessor
from ..proba_dists import ProbaDist
from .base_func import BaseFunc, ExampleData, Inputs, ArgsType1, ArgsType2, ModelTypes


__all__ = (
    'TransitionModel',
)


class TransitionModel(BaseFunc):
    r"""

    A deterministic transition function :math:`s'_\theta(s,a)`.

    Parameters
    ----------
    func : function

        A Haiku-style function that specifies the forward pass. The function signature must be the
        same as the example below.

    env : gym.Env

        The gym-style environment. This is used to validate the input/output structure of ``func``.

    observation_preprocessor : function, optional

        Turns a single observation into a batch of observations in a form that is convenient for
        feeding into :code:`func`. If left unspecified, this defaults to
        :attr:`proba_dist.preprocess_variate <coax.proba_dists.ProbaDist.preprocess_variate>`. The
        reason why the default is not :func:`coax.utils.default_preprocessor` is that we prefer
        consistence with :class:`coax.StochasticTransitionModel`.

    observation_postprocessor : function, optional

        Takes a batch of generated observations and makes sure that they are that are compatible
        with the original :code:`observation_space`. If left unspecified, this defaults to
        :attr:`proba_dist.postprocess_variate <coax.proba_dists.ProbaDist.postprocess_variate>`.

    action_preprocessor : function, optional

        Turns a single action into a batch of actions in a form that is convenient for feeding into
        :code:`func`. If left unspecified, this defaults
        :func:`default_preprocessor(env.action_space) <coax.utils.default_preprocessor>`.

    random_seed : int, optional

        Seed for pseudo-random number generators.

    """
    def __init__(
            self, func, env, observation_preprocessor=None, observation_postprocessor=None,
            action_preprocessor=None, random_seed=None):

        self.observation_preprocessor = observation_preprocessor
        self.observation_postprocessor = observation_postprocessor
        self.action_preprocessor = action_preprocessor

        # defaults
        if self.observation_preprocessor is None:
            self.observation_preprocessor = ProbaDist(env.observation_space).preprocess_variate
        if self.observation_postprocessor is None:
            self.observation_postprocessor = ProbaDist(env.observation_space).postprocess_variate
        if self.action_preprocessor is None:
            self.action_preprocessor = default_preprocessor(env.action_space)

        super().__init__(
            func,
            observation_space=env.observation_space,
            action_space=env.action_space,
            random_seed=random_seed)

    def __call__(self, s, a=None):
        r"""

        Evaluate the state-action function on a state observation :math:`s` or
        on a state-action pair :math:`(s, a)`.

        Parameters
        ----------
        s : state observation

            A single state observation :math:`s`.

        a : action

            A single action :math:`a`.

        Returns
        -------
        q_sa or q_s : ndarray

            Depending on whether :code:`a` is provided, this either returns a scalar representing
            :math:`q(s,a)\in\mathbb{R}` or a vector representing :math:`q(s,.)\in\mathbb{R}^n`,
            where :math:`n` is the number of discrete actions. Naturally, this only applies for
            discrete action spaces.

        """
        S = self.observation_preprocessor(self.rng, s)
        if a is None:
            S_next, _ = self.function_type2(self.params, self.function_state, self.rng, S, False)
            S_next = batch_to_single(S_next)  # (batch, num_actions, *) -> (num_actions, *)
            n = self.action_space.n
            s_next = [self.observation_postprocessor(self.rng, S_next, index=i) for i in range(n)]
        else:
            A = self.action_preprocessor(self.rng, a)
            S_next, _ = self.function_type1(self.params, self.function_state, self.rng, S, A, False)
            s_next = self.observation_postprocessor(self.rng, S_next)
        return s_next

    @property
    def function_type1(self):
        r"""

        Same as :attr:`function`, except that it ensures a type-1 function signature, regardless of
        the underlying :attr:`modeltype`.

        """
        if self.modeltype == 1:
            return self.function

        assert isinstance(self.action_space, Discrete)

        def project(A):
            assert A.ndim == 2, f"bad shape: {A.shape}"
            assert A.shape[1] == self.action_space.n, f"bad shape: {A.shape}"
            def func(leaf):  # noqa: E306
                assert isinstance(leaf, jnp.ndarray), f"leaf must be ndarray, got: {type(leaf)}"
                assert leaf.ndim >= 2, f"bad shape: {leaf.shape}"
                assert leaf.shape[0] == A.shape[0], \
                    f"batch_size (axis=0) mismatch: leaf.shape: {leaf.shape}, A.shape: {A.shape}"
                assert leaf.shape[1] == A.shape[1], \
                    f"num_actions (axis=1) mismatch: leaf.shape: {leaf.shape}, A.shape: {A.shape}"
                return jax.vmap(jnp.dot)(jnp.moveaxis(leaf, 1, -1), A)
            return func

        def type1_func(type2_params, type2_state, rng, S, A, is_training):
            S_next, state_new = self.function(type2_params, type2_state, rng, S, is_training)
            S_next = jax.tree_map(project(A), S_next)
            return S_next, state_new

        return type1_func

    @property
    def function_type2(self):
        r"""

        Same as :attr:`function`, except that it ensures a type-2 function signature, regardless of
        the underlying :attr:`modeltype`.

        """
        if self.modeltype == 2:
            return self.function

        if not isinstance(self.action_space, Discrete):
            raise ValueError(
                "input 'A' is required for type-1 dynamics model when action space is non-Discrete")

        n = self.action_space.n

        def reshape(leaf):
            # reshape from (batch * num_actions, *shape) -> (batch, *shape, num_actions)
            assert isinstance(leaf, jnp.ndarray), f"all leaves must be ndarray, got: {type(leaf)}"
            assert leaf.ndim >= 1, f"bad shape: {leaf.shape}"
            assert leaf.shape[0] % n == 0, \
                f"first axis size must be a multiple of num_actions, got shape: {leaf.shape}"
            leaf = jnp.reshape(leaf, (-1, n, *leaf.shape[1:]))  # (batch, num_actions, *shape)
            return leaf

        def type2_func(type1_params, type1_state, rng, S, is_training):
            rngs = hk.PRNGSequence(rng)
            batch_size = jax.tree_leaves(S)[0].shape[0]

            # example: let S = [7, 2, 5, 8] and num_actions = 3, then
            # S_rep = [7, 7, 7, 2, 2, 2, 5, 5, 5, 8, 8, 8]  # repeated
            # A_rep = [0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2]  # tiled
            S_rep = jax.tree_map(lambda x: jnp.repeat(x, n, axis=0), S)
            A_rep = jnp.tile(jnp.arange(n), batch_size)
            A_rep = self.action_preprocessor(next(rngs), A_rep)  # one-hot encoding

            # evaluate on replicas => output shape: (batch * num_actions, *shape)
            S_next_rep, state_new = self.function(
                type1_params, type1_state, next(rngs), S_rep, A_rep, is_training)
            S_next = jax.tree_map(reshape, S_next_rep)

            return S_next, state_new

        return type2_func

    @property
    def modeltype(self):
        r"""

        Specifier for how the transition function is modeled, i.e.

        .. math::

            (s,a)   &\mapsto s'(s,a)    &\qquad (\text{modeltype} &= 1) \\
            s       &\mapsto s'(s,.)    &\qquad (\text{modeltype} &= 2)

        Note that modeltype=2 is only well-defined if the action space is :class:`Discrete
        <gym.spaces.Discrete>`. Namely, :math:`n` is the number of discrete actions.

        """
        return self._modeltype

    @classmethod
    def example_data(
            cls, env, observation_preprocessor=None, action_preprocessor=None,
            batch_size=1, random_seed=None):

        if not isinstance(env.observation_space, Space):
            raise TypeError(
                "env.observation_space must be derived from gym.Space, "
                f"got: {type(env.observation_space)}")

        if not isinstance(env.action_space, Space):
            raise TypeError(
                f"env.action_space must be derived from gym.Space, got: {type(env.action_space)}")

        if observation_preprocessor is None:
            observation_preprocessor = ProbaDist(env.observation_space).preprocess_variate

        if action_preprocessor is None:
            action_preprocessor = default_preprocessor(env.action_space)

        rnd = onp.random.RandomState(random_seed)
        rngs = hk.PRNGSequence(rnd.randint(jnp.iinfo('int32').max))

        # input: state observations
        S = [safe_sample(env.observation_space, rnd) for _ in range(batch_size)]
        S = [observation_preprocessor(next(rngs), s) for s in S]
        S = jax.tree_map(lambda *x: jnp.concatenate(x, axis=0), *S)

        # input: actions
        A = [safe_sample(env.action_space, rnd) for _ in range(batch_size)]
        A = [action_preprocessor(next(rngs), a) for a in A]
        A = jax.tree_map(lambda *x: jnp.concatenate(x, axis=0), *A)

        # output: type1
        S_next_type1 = jax.tree_map(lambda x: jnp.asarray(rnd.randn(batch_size, *x.shape[1:])), S)
        q1_data = ExampleData(
            inputs=Inputs(args=ArgsType1(S=S, A=A, is_training=True), static_argnums=(2,)),
            output=S_next_type1)

        if not isinstance(env.action_space, Discrete):
            return ModelTypes(type1=q1_data, type2=None)

        # output: type2 (if actions are discrete)
        S_next_type2 = jax.tree_map(
            lambda x: jnp.asarray(rnd.randn(batch_size, env.action_space.n, *x.shape[1:])), S)
        q2_data = ExampleData(
            inputs=Inputs(args=ArgsType2(S=S, is_training=True), static_argnums=(1,)),
            output=S_next_type2)

        return ModelTypes(type1=q1_data, type2=q2_data)

    def _check_signature(self, func):
        sig_type1 = ('S', 'A', 'is_training')
        sig_type2 = ('S', 'is_training')
        sig = tuple(signature(func).parameters)

        if sig not in (sig_type1, sig_type2):
            sig = ', '.join(sig)
            alt = ' or func(S, is_training)' if isinstance(self.action_space, Discrete) else ''
            raise TypeError(
                f"func has bad signature; expected: func(S, A, is_training){alt}, got: func({sig})")

        if sig == sig_type2 and not isinstance(self.action_space, Discrete):
            raise TypeError("type-2 models are only well-defined for Discrete action spaces")

        Env = namedtuple('Env', ('observation_space', 'action_space'))
        example_data_per_modeltype = self.example_data(
            env=Env(self.observation_space, self.action_space),
            action_preprocessor=self.action_preprocessor,
            batch_size=1,
            random_seed=self.random_seed)

        if sig == sig_type1:
            self._modeltype = 1
            example_data = example_data_per_modeltype.type1
        else:
            self._modeltype = 2
            example_data = example_data_per_modeltype.type2

        return example_data

    def _check_output(self, actual, expected):
        expected_leaves, expected_structure = jax.tree_flatten(expected)
        actual_leaves, actual_structure = jax.tree_flatten(actual)
        assert all(isinstance(x, jnp.ndarray) for x in expected_leaves), "bad example_data"

        if actual_structure != expected_structure:
            raise TypeError(
                f"func has bad return tree_structure, expected: {expected_structure}, "
                f"got: {actual_structure}")

        if not all(isinstance(x, jnp.ndarray) for x in actual_leaves):
            bad_types = tuple(type(x) for x in actual_leaves if not isinstance(x, jnp.ndarray))
            raise TypeError(
                "all leaves of dist_params must be of type: jax.numpy.ndarray, "
                f"found leaves of type: {bad_types}")

        if not all(a.shape == b.shape for a, b in zip(actual_leaves, expected_leaves)):
            shapes_tree = jax.tree_map(
                lambda a, b: f"{a.shape} {'!=' if a.shape != b.shape else '=='} {b.shape}",
                actual, expected)
            raise TypeError(f"found leaves with unexpected shapes: {shapes_tree}")



import haiku as hk
import jax
import jax.numpy as jnp
import numpy as onp

__all__ = (
    'quantiles',
    'quantiles_uniform',
    'quantile_cos_embedding'
)


def quantiles_uniform(rng, batch_size, num_quantiles=32):
    """
    Generate :code:`batch_size` quantile fractions that split the interval :math:`[0, 1]`
    into :code:`num_quantiles` uniformly distributed fractions.

    Parameters
    ----------
    rng : jax.random.PRNGKey
        A pseudo-random number generator key.
    batch_size : int
        The batch size for which the quantile fractions should be generated.
    num_quantiles : int, optional
        The number of quantile fractions. By default 32.

    Returns
    -------
    quantile_fractions : ndarray
        Array of quantile fractions.
    """
    rngs = hk.PRNGSequence(rng)
    quantile_fractions = jax.random.uniform(next(rngs), shape=(batch_size, num_quantiles))
    quantile_fraction_differences = quantile_fractions / \
        jnp.sum(quantile_fractions, axis=-1, keepdims=True)
    quantile_fractions = jnp.cumsum(quantile_fraction_differences, axis=-1)
    return quantile_fractions


def quantiles(batch_size, num_quantiles=200):
    r"""
    Generate :code:`batch_size` quantile fractions that split the interval :math:`[0, 1]`
    into :code:`num_quantiles` equally spaced fractions.

    Parameters
    ----------
    batch_size : int
        The batch size for which the quantile fractions should be generated.
    num_quantiles : int, optional
        The number of quantile fractions. By default 200.

    Returns
    -------
    quantile_fractions : ndarray

        Array of quantile fractions.
    """
    quantile_fractions = jnp.arange(num_quantiles, dtype=jnp.float32) / num_quantiles
    quantile_fractions = jnp.tile(quantile_fractions[None, :], [batch_size, 1])
    return quantile_fractions


def quantile_cos_embedding(quantile_fractions, n=64):
    r"""
    Embed the given quantile fractions :math:`\tau` in an `n` dimensional space
    using cosine basis functions.

    .. math::

        \phi(\tau) = \cos(\tau i \pi) \qquad 0 \leq i \lt n


    Parameters
    ----------
    quantile_fractions : ndarray

        Array of quantile fractions :math:`\tau` to be embedded.

    n : int

        The dimensionality of the embedding. By default 64.

    Returns
    -------
    quantile_embs : ndarray

        Array of quantile embeddings with shape `(quantile_fractions.shape[0], n)`.
    """
    quantile_fractions = jnp.tile(quantile_fractions[..., None],
                                  [1] * quantile_fractions.ndim + [n])
    quantiles_emb = (
        jnp.arange(1, n + 1, 1)
        * onp.pi
        * quantile_fractions)
    quantiles_emb = jnp.cos(quantiles_emb)
    return quantiles_emb



from collections import deque

import gym


class FrameStacking(gym.Wrapper):
    r"""

    Wrapper that does frame stacking (see `DQN paper
    <https://www.cs.toronto.edu/~vmnih/docs/dqn.pdf>`_).

    This implementation is different from most implementations in that it doesn't perform the
    stacking itself. Instead, it just returns a tuple of frames (untouched), which may be stacked
    downstream.

    The benefit of this implementation is two-fold. First, it respects the :mod:`gym.spaces` API,
    where each observation is truly an element of the observation space (this is not true of the gym
    implementation, which uses a custom data class to maintain its minimal memory footprint).
    Second, this implementation is compatibility with the :mod:`jax.tree_util` module, which means
    that we can feed it into jit-compiled functions directly.

    Example
    -------

    .. code::

        import gym
        env = gym.make('PongNoFrameskip-v0')
        print(env.observation_space)  # Box(210, 160, 3)

        env = FrameStacking(env, num_frames=2)
        print(env.observation_space)  # Tuple((Box(210, 160, 3), Box(210, 160, 3)))

    Parameters
    ----------
    env : gym-style environment

        The original environment to be wrapped.

    num_frames : positive int

        Number of frames to stack.

    """
    def __init__(self, env, num_frames):
        if not (isinstance(num_frames, int) and num_frames > 0):
            raise TypeError(f"num_frames must be a positive int, got: {num_frames}")

        super().__init__(env)
        self.observation_space = gym.spaces.Tuple((self.env.observation_space,) * num_frames)
        self._frames = deque(maxlen=num_frames)

    def step(self, action):
        observation, reward, done, info = self.env.step(action)
        self._frames.append(observation)
        return tuple(self._frames), reward, done, info

    def reset(self, **kwargs):
        observation = self.env.reset(**kwargs)
        self._frames.extend(observation for _ in range(self._frames.maxlen))
        return tuple(self._frames)  # shallow copy



import os
import re
import datetime
import time
from collections import deque
from typing import Mapping

import numpy as np
import lz4.frame
import cloudpickle as pickle
from gym import Wrapper
from gym.spaces import Discrete
from tensorboardX import SummaryWriter

from .._base.mixins import LoggerMixin
from ..utils import enable_logging


__all__ = (
    'TrainMonitor',
)


class StreamingSample:
    def __init__(self, maxlen, random_seed=None):
        self._deque = deque(maxlen=maxlen)
        self._count = 0
        self._rnd = np.random.RandomState(random_seed)

    def reset(self):
        self._deque = deque(maxlen=self.maxlen)
        self._count = 0

    def append(self, obj):
        self._count += 1
        if len(self) < self.maxlen:
            self._deque.append(obj)
        elif self._rnd.rand() < self.maxlen / self._count:
            i = self._rnd.randint(self.maxlen)
            self._deque[i] = obj

    @property
    def values(self):
        return list(self._deque)  # shallow copy

    @property
    def maxlen(self):
        return self._deque.maxlen

    def __len__(self):
        return len(self._deque)

    def __bool__(self):
        return bool(self._deque)


class TrainMonitor(Wrapper, LoggerMixin):
    r"""
    Environment wrapper for monitoring the training process.

    This wrapper logs some diagnostics at the end of each episode and it also gives us some handy
    attributes (listed below).

    Parameters
    ----------
    env : gym environment

        A gym environment.

    tensorboard_dir : str, optional

        If provided, TrainMonitor will log all diagnostics to be viewed in tensorboard. To view
        these, point tensorboard to the same dir:

        .. code:: bash

            $ tensorboard --logdir {tensorboard_dir}

    tensorboard_write_all : bool, optional

        You may record your training metrics using the :attr:`record_metrics` method. Setting the
        ``tensorboard_write_all`` specifies whether to pass the metrics on to tensorboard
        immediately (``True``) or to wait and average them across the episode (``False``). The
        default setting (``False``) prevents tensorboard from being fluided by logs.

    log_all_metrics : bool, optional

        Whether to log all metrics. If ``log_all_metrics=False``, only a reduced set of metrics are
        logged.

    smoothing : positive int, optional

        The number of observations for smoothing the metrics. We use the following smooth update
        rule:

        .. math::

            n\ &\leftarrow\ \min(\text{smoothing}, n + 1) \\
            x_\text{avg}\ &\leftarrow\ x_\text{avg}
                + \frac{x_\text{obs} - x_\text{avg}}{n}

    \*\*logger_kwargs

        Keyword arguments to pass on to :func:`coax.utils.enable_logging`.

    Attributes
    ----------
    T : positive int

        Global step counter. This is not reset by ``env.reset()``, use ``env.reset_global()``
        instead.

    ep : positive int

        Global episode counter. This is not reset by ``env.reset()``, use ``env.reset_global()``
        instead.

    t : positive int

        Step counter within an episode.

    G : float

        The return, i.e. amount of reward accumulated from the start of the current episode.

    avg_G : float

        The average return G, averaged over the past 100 episodes.

    dt_ms : float

        The average wall time of a single step, in milliseconds.

    """
    _COUNTER_ATTRS = (
        'T', 'ep', 't', 'G', 'avg_G', '_n_avg_G', '_ep_starttime', '_ep_metrics', '_ep_actions',
        '_tensorboard_dir', '_period')

    def __init__(
            self, env,
            tensorboard_dir=None,
            tensorboard_write_all=False,
            log_all_metrics=False,
            smoothing=10,
            **logger_kwargs):

        super().__init__(env)
        self.log_all_metrics = log_all_metrics
        self.tensorboard_write_all = tensorboard_write_all
        self.smoothing = float(smoothing)
        self.reset_global()
        enable_logging(**logger_kwargs)
        self._init_tensorboard(tensorboard_dir)

    def reset_global(self):
        r""" Reset the global counters, not just the episodic ones. """
        self.T = 0
        self.ep = 0
        self.t = 0
        self.G = 0.0
        self.avg_G = 0.0
        self._n_avg_G = 0.0
        self._ep_starttime = time.time()
        self._ep_metrics = {}
        self._ep_actions = StreamingSample(maxlen=1000)
        self._period = {'T': {}, 'ep': {}}

    def reset(self):
        # write logs from previous episode:
        if self.ep:
            self._write_episode_logs()

        # increment global counters:
        self.T += 1
        self.ep += 1

        # reset episodic counters:
        self.t = 0
        self.G = 0.0
        self._ep_starttime = time.time()
        self._ep_metrics = {}
        self._ep_actions.reset()

        return self.env.reset()

    @property
    def dt_ms(self):
        if self.t <= 0:
            return np.nan
        return 1000 * (time.time() - self._ep_starttime) / self.t

    @property
    def avg_r(self):
        if self.t <= 0:
            return np.nan
        return self.G / self.t

    def step(self, a):
        self._ep_actions.append(a)
        s_next, r, done, info = self.env.step(a)
        if info is None:
            info = {}
        info['monitor'] = {'T': self.T, 'ep': self.ep}
        self.t += 1
        self.T += 1
        self.G += r
        if done:
            if self._n_avg_G < self.smoothing:
                self._n_avg_G += 1.
            self.avg_G += (self.G - self.avg_G) / self._n_avg_G

        return s_next, r, done, info

    def record_metrics(self, metrics):
        r"""
        Record metrics during the training process.

        These are used to print more diagnostics.

        Parameters
        ----------
        metrics : dict

            A dict of metrics, of type ``{name <str>: value <float>}``.

        """
        if not isinstance(metrics, Mapping):
            raise TypeError("metrics must be a Mapping")

        # write metrics to tensoboard
        if self.tensorboard is not None and self.tensorboard_write_all:
            for name, metric in metrics.items():
                self.tensorboard.add_scalar(
                    str(name), float(metric), global_step=self.T)

        # compute episode averages
        for k, v in metrics.items():
            if k not in self._ep_metrics:
                self._ep_metrics[k] = v, 1.
            else:
                x, n = self._ep_metrics[k]
                self._ep_metrics[k] = x + v, n + 1

    def get_metrics(self):
        r"""
        Return the current state of the metrics.

        Returns
        -------
        metrics : dict

            A dict of metrics, of type ``{name <str>: value <float>}``.

        """
        return {k: float(x) / n for k, (x, n) in self._ep_metrics.items()}

    def period(self, name, T_period=None, ep_period=None):
        if T_period is not None:
            T_period = int(T_period)
            assert T_period > 0
            if name not in self._period['T']:
                self._period['T'][name] = 1
            if self.T >= self._period['T'][name] * T_period:
                self._period['T'][name] += 1
                return True or self.period(name, None, ep_period)
            return self.period(name, None, ep_period)
        if ep_period is not None:
            ep_period = int(ep_period)
            assert ep_period > 0
            if name not in self._period['ep']:
                self._period['ep'][name] = 1
            if self.ep >= self._period['ep'][name] * ep_period:
                self._period['ep'][name] += 1
                return True
        return False

    @property
    def tensorboard(self):
        if not hasattr(self, '_tensorboard'):
            assert self._tensorboard_dir is not None
            self._tensorboard = SummaryWriter(self._tensorboard_dir)
        return self._tensorboard

    def _init_tensorboard(self, tensorboard_dir):
        if tensorboard_dir is None:
            self._tensorboard_dir = None
            self._tensorboard = None
            return

        # append timestamp to disambiguate instances
        if not re.match(r'.*/\d{8}_\d{6}$', tensorboard_dir):
            tensorboard_dir = os.path.join(
                tensorboard_dir,
                datetime.datetime.now().strftime('%Y%m%d_%H%M%S'))

        # only set/update if necessary
        if tensorboard_dir != getattr(self, '_tensorboard_dir', None):
            self._tensorboard_dir = tensorboard_dir
            if hasattr(self, '_tensorboard'):
                del self._tensorboard

    def _write_episode_logs(self):
        metrics = (
            f'{k:s}: {float(x) / n:.3g}'
            for k, (x, n) in self._ep_metrics.items() if (
                self.log_all_metrics
                or str(k).endswith('/loss')
                or str(k).endswith('/entropy')
                or str(k).endswith('/kl_div')
                or str(k).startswith('throughput/')
            )
        )
        self.logger.info(
            ',\t'.join((
                f'ep: {self.ep:d}',
                f'T: {self.T:,d}',
                f'G: {self.G:.3g}',
                f'avg_r: {self.avg_r:.3g}',
                f'avg_G: {self.avg_G:.3g}',
                f't: {self.t:d}',
                f'dt: {self.dt_ms:.3f}ms',
                *metrics)))

        if self.tensorboard is not None:
            metrics = {
                'episode/episode': self.ep,
                'episode/avg_reward': self.avg_r,
                'episode/return': self.G,
                'episode/steps': self.t,
                'episode/avg_step_duration_ms': self.dt_ms}
            for name, metric in metrics.items():
                self.tensorboard.add_scalar(
                    str(name), float(metric), global_step=self.T)
            if self._ep_actions:
                if isinstance(self.action_space, Discrete):
                    bins = np.arange(self.action_space.n + 1)
                else:
                    bins = 'auto'  # see also: np.histogram_bin_edges.__doc__
                self.tensorboard.add_histogram(
                    tag='actions', values=self._ep_actions.values, global_step=self.T, bins=bins)
            if self._ep_metrics and not self.tensorboard_write_all:
                for k, (x, n) in self._ep_metrics.items():
                    self.tensorboard.add_scalar(str(k), float(x) / n, global_step=self.T)
            self.tensorboard.flush()

    def __getstate__(self):
        state = self.__dict__.copy()   # shallow copy
        if '_tensorboard' in state:
            del state['_tensorboard']  # remove reference to non-pickleable attr
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._init_tensorboard(state['_tensorboard_dir'])

    def get_counters(self):
        r"""

        Get the current state of all internal counters.

        Returns
        -------
        counter : dict

            The dict that contains the counters.

        """
        return {k: getattr(self, k) for k in self._COUNTER_ATTRS}

    def set_counters(self, counters):
        r"""

        Restore the state of all internal counters.

        Parameters
        ----------
        counter : dict

            The dict that contains the counters.

        """
        if not (isinstance(counters, dict) and set(counters) == set(self._COUNTER_ATTRS)):
            raise TypeError(f"invalid counters dict: {counters}")
        self.__setstate__(counters)

    def save_counters(self, filepath):
        r"""

        Store the current state of all internal counters.

        Parameters
        ----------
        filepath : str

            The checkpoint file path.

        """
        counters = self.get_counters()
        os.makedirs(os.path.dirname(filepath) or '.', exist_ok=True)

        with lz4.frame.open(filepath, 'wb') as f:
            f.write(pickle.dumps(counters))

    def load_counters(self, filepath):
        r"""

        Restore the state of all internal counters.

        Parameters
        ----------
        filepath : str

            The checkpoint file path.

        """
        with lz4.frame.open(filepath, 'rb') as f:
            counters = pickle.loads(f.read())
        self.set_counters(counters)



from abc import ABC, abstractmethod


__all__ = (
    'BaseReplayBuffer',
)


class BaseReplayBuffer(ABC):

    @property
    @abstractmethod
    def capacity(self):
        pass

    @abstractmethod
    def add(self, transition_batch):
        pass

    @abstractmethod
    def sample(self, batch_size=32):
        pass

    @abstractmethod
    def clear(self):
        pass

    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def __bool__(self):
        pass

    @abstractmethod
    def __iter__(self):
        pass



from abc import ABC, abstractmethod
from typing import Any, Tuple, NamedTuple

import jax
import haiku as hk
from gym.spaces import Space

from ..typing import Batch, Observation, Action
from ..utils import pretty_repr, jit
from .._base.mixins import RandomStateMixin, CopyMixin


class Inputs(NamedTuple):
    args: Any
    static_argnums: Tuple[int, ...]

    def __repr__(self):
        return pretty_repr(self)


class ExampleData(NamedTuple):
    inputs: Inputs
    output: Any

    def __repr__(self):
        return pretty_repr(self)


class ArgsType1(NamedTuple):
    S: Batch[Observation]
    A: Batch[Action]
    is_training: bool

    def __repr__(self):
        return pretty_repr(self)


class ArgsType2(NamedTuple):
    S: Batch[Observation]
    is_training: bool

    def __repr__(self):
        return pretty_repr(self)


class ModelTypes(NamedTuple):
    type1: ArgsType1
    type2: ArgsType2

    def __repr__(self):
        return pretty_repr(self)


class BaseFunc(ABC, RandomStateMixin, CopyMixin):
    """ Abstract base class for function approximators: coax.V, coax.Q, coax.Policy """

    def __init__(self, func, observation_space, action_space=None, random_seed=None):

        if not isinstance(observation_space, Space):
            raise TypeError(
                f"observation_space must be derived from gym.Space, got: {type(observation_space)}")
        self.observation_space = observation_space

        if action_space is not None:
            if not isinstance(action_space, Space):
                raise TypeError(
                    f"action_space must be derived from gym.Space, got: {type(action_space)}")
            self.action_space = action_space

        self.random_seed = random_seed  # also initializes self.rng via RandomStateMixin
        self._jitted_funcs = {}

        # Haiku-transform the provided func
        example_data = self._check_signature(func)
        static_argnums = tuple(i + 3 for i in example_data.inputs.static_argnums)
        transformed = hk.transform_with_state(func)
        self._function = jit(transformed.apply, static_argnums=static_argnums)

        # init function params and state
        self._params, self._function_state = transformed.init(self.rng, *example_data.inputs.args)

        # check if output has the expected shape etc.
        output, _ = \
            self._function(self.params, self.function_state, self.rng, *example_data.inputs.args)
        self._check_output(output, example_data.output)

        def soft_update_func(old, new, tau):
            return jax.tree_map(lambda a, b: (1 - tau) * a + tau * b, old, new)

        self._soft_update_func = jit(soft_update_func)

    def soft_update(self, other, tau):
        r""" Synchronize the current instance with ``other`` through exponential smoothing:

        .. math::

            \theta\ \leftarrow\ \theta + \tau\, (\theta_\text{new} - \theta)

        Parameters
        ----------
        other

            A seperate copy of the current object. This object will hold the new parameters
            :math:`\theta_\text{new}`.

        tau : float between 0 and 1, optional

            If we set :math:`\tau=1` we do a hard update. If we pick a smaller value, we do a smooth
            update.

        """
        if not isinstance(other, self.__class__):
            raise TypeError("'self' and 'other' must be of the same type")

        self.params = self._soft_update_func(self.params, other.params, tau)
        self.function_state = self._soft_update_func(self.function_state, other.function_state, tau)

    @property
    def params(self):
        """ The parameters (weights) of the function approximator. """
        return self._params

    @params.setter
    def params(self, new_params):
        if jax.tree_structure(new_params) != jax.tree_structure(self._params):
            raise TypeError("new params must have the same structure as old params")
        self._params = new_params

    @property
    def function(self):
        r"""

        The function approximator itself, defined as a JIT-compiled pure function. This function may
        be called directly as:

        .. code:: python

            output, function_state = obj.function(obj.params, obj.function_state, obj.rng, *inputs)

        """
        return self._function

    @property
    def function_state(self):
        """ The state of the function approximator, see :func:`haiku.transform_with_state`. """
        return self._function_state

    @function_state.setter
    def function_state(self, new_function_state):
        if jax.tree_structure(new_function_state) != jax.tree_structure(self._function_state):
            raise TypeError("new function_state must have the same structure as old function_state")
        self._function_state = new_function_state

    @abstractmethod
    def _check_signature(self, func):
        """ Check if func has expected input signature; returns example_data; raises TypeError """

    @abstractmethod
    def _check_output(self, actual, expected):
        """ Check if func has expected output signature; raises TypeError """

    @abstractmethod
    def example_data(self, env, **kwargs):
        r"""

        A small utility function that generates example input and output data. These may be useful
        for writing and debugging your own custom function approximators.

        """



r"""
.. autosummary::
    :nosignatures:

    coax.proba_dists.ProbaDist
    coax.proba_dists.CategoricalDist
    coax.proba_dists.NormalDist
    coax.proba_dists.DiscretizedIntervalDist
    coax.proba_dists.EmpiricalQuantileDist

-----

Probability Distributions
=========================

This is a collection of **differentiable** probability distributions used throughout the package.


Object Reference
----------------

.. autoclass:: coax.proba_dists.ProbaDist
.. autoclass:: coax.proba_dists.CategoricalDist
.. autoclass:: coax.proba_dists.NormalDist
.. autoclass:: coax.proba_dists.DiscretizedIntervalDist
.. autoclass:: coax.proba_dists.EmpiricalQuantileDist


"""

from ._composite import ProbaDist
from ._categorical import CategoricalDist
from ._normal import NormalDist
from ._discretized_interval import DiscretizedIntervalDist
from ._empirical_quantile import EmpiricalQuantileDist


__all__ = (
    'ProbaDist',
    'CategoricalDist',
    'NormalDist',
    'DiscretizedIntervalDist',
    'EmpiricalQuantileDist'
)



import haiku as hk

from ..utils import is_stochastic, jit
from .._core.base_stochastic_func_type1 import BaseStochasticFuncType1
from .._core.base_stochastic_func_type2 import BaseStochasticFuncType2


class Regularizer:
    r"""

    Abstract base class for policy regularizers. Check out
    :class:`coax.regularizers.EntropyRegularizer` for a specific example.

    Parameters
    ----------
    f : stochastic function approximator

        The stochastic function approximator (e.g. :class:`coax.Policy`) to regularize.

    """
    def __init__(self, f):
        if not is_stochastic(f):
            raise TypeError(f"proba_dist must be a stochastic function, got {type(f)}")
        self.f = f

    @property
    def hyperparams(self):
        return {}

    @property
    def function(self):
        r"""

        JIT-compiled function that returns the values for the regularization term.

        Parameters
        ----------
        dist_params : pytree with ndarray leaves

            The distribution parameters of the (conditional) probability distribution.

        \*\*hyperparams

            Hyperparameters specific to the regularizer, see :attr:`hyperparams`.

        """
        return self._function

    @property
    def metrics_func(self):
        r"""

        JIT-compiled function that returns the performance metrics for the regularization term.

        Parameters
        ----------
        dist_params : pytree with ndarray leaves

            The distribution parameters of the (conditional) probability distribution
            :math:`\pi(a|s)`.

        \*\*hyperparams

            Hyperparameters specific to the regularizer, see :attr:`hyperparams`.

        """
        return self._metrics_func

    @property
    def batch_eval(self):
        if not hasattr(self, '_batch_eval_func'):
            def batch_eval_func(params, hyperparams, state, rng, transition_batch):
                rngs = hk.PRNGSequence(rng)
                if isinstance(self.f, BaseStochasticFuncType1):
                    S = self.f.observation_preprocessor(next(rngs), transition_batch.S)
                    A = self.f.action_preprocessor(next(rngs), transition_batch.A)
                    dist_params, _ = self.f.function(params, state, next(rngs), S, A, False)
                if isinstance(self.f, BaseStochasticFuncType2):
                    S = self.f.observation_preprocessor(next(rngs), transition_batch.S)
                    dist_params, _ = self.f.function(params, state, next(rngs), S, False)
                else:
                    raise TypeError(
                        "f must be derived from BaseStochasticFuncType1 or BaseStochasticFuncType2")
                return self.function(dist_params, **hyperparams), self.metrics_func(dist_params,
                                                                                    **hyperparams)

            self._batch_eval_func = jit(batch_eval_func)

        return self._batch_eval_func



import time
import inspect
from abc import ABC, abstractmethod
from typing import Optional

import gym
from jax.lib.xla_bridge import get_backend

from ..typing import Policy
from ..wrappers import TrainMonitor
from ..reward_tracing._base import BaseRewardTracer
from ..experience_replay._base import BaseReplayBuffer


__all__ = (
    'Worker',
)


class WorkerError(Exception):
    pass


class Worker(ABC):
    r"""

    The base class for defining workers as part of a distributed agent.

    Parameters
    ----------
    env : gym.Env | str | function

        Specifies the gym-style environment by either passing the env itself (gym.Env), its name
        (str), or a function that generates the environment.

    param_store : Worker, optional

        A distributed agent is presumed to have one worker that plays the role of a parameter store.
        To define the parameter-store worker itself, you must leave :code:`param_store=None`. For
        other worker roles, however, :code:`param_store` must be provided.

    pi : Policy, optional

        The behavior policy that is used by rollout workers to generate experience.

    tracer : RewardTracer, optional

        The reward tracer that is used by rollout workers.

    buffer : ReplayBuffer, optional

        The experience-replay buffer that is populated by rollout workers and sampled from by
        learners.

    buffer_warmup : int, optional

        The warmup period for the experience replay buffer, i.e. the minimal number of transitions
        that need to be stored in the replay buffer before we start sampling from it.

    name : str, optional

        A human-readable identifier of the worker.

    """
    pi: Optional[Policy] = None
    tracer: Optional[BaseRewardTracer] = None
    buffer: Optional[BaseReplayBuffer] = None
    buffer_warmup: Optional[int] = None

    def __init__(
            self, env,
            param_store=None,
            pi=None,
            tracer=None,
            buffer=None,
            buffer_warmup=None,
            name=None):

        # import inline to avoid hard dependency on ray
        import ray
        import ray.actor
        self.__ray = ray

        self.env = _check_env(env, name)
        self.param_store = param_store
        self.pi = pi
        self.tracer = tracer
        self.buffer = buffer
        self.buffer_warmup = buffer_warmup
        self.name = name
        self.env.logger.info(f"JAX platform name: '{get_backend().platform}'")

    @abstractmethod
    def get_state(self):
        r"""

        Get the internal state that is shared between workers.

        Returns
        -------
        state : object

            The internal state. This will be consumed by :func:`set_state(state) <set_state>`.

        """
        pass

    @abstractmethod
    def set_state(self, state):
        r"""

        Set the internal state that is shared between workers.

        Parameters
        ----------
        state : object

            The internal state, as returned by :func:`get_state`.

        """
        pass

    @abstractmethod
    def trace(self, s, a, r, done, logp=0.0, w=1.0):
        r"""

        This implements the reward-tracing step of a single, raw transition.

        Parameters
        ----------
        s : state observation

            A single state observation.

        a : action

            A single action.

        r : float

            A single observed reward.

        done : bool

            Whether the episode has finished.

        logp : float, optional

            The log-propensity :math:`\log\pi(a|s)`.

        w : float, optional

            Sample weight associated with the given state-action pair.


        """
        pass

    @abstractmethod
    def learn(self, transition_batch):
        r""" Update the model parameters given a transition batch. """
        pass

    def rollout(self):
        assert self.pi is not None
        s = self.env.reset()
        for t in range(self.env.spec.max_episode_steps):
            a, logp = self.pi(s, return_logp=True)
            s_next, r, done, info = self.env.step(a)

            self.trace(s, a, r, done, logp)

            if done:
                break

            s = s_next

    def rollout_loop(self, max_total_steps, reward_threshold=None):
        reward_threshold = _check_reward_threshold(reward_threshold, self.env)
        T_global = self.pull_getattr('env.T')
        while T_global < max_total_steps and self.env.avg_G < reward_threshold:
            self.pull_state()
            self.rollout()
            metrics = self.pull_metrics()
            metrics['throughput/rollout_loop'] = 1000 / self.env.dt_ms
            metrics['episode/T_global'] = T_global = self.pull_getattr('env.T') + self.env.t
            self.push_setattr('env.T', T_global)  # not exactly thread-safe, but that's okay
            self.env.record_metrics(metrics)

    def learn_loop(self, max_total_steps, batch_size=32):
        throughput = 0.
        while self.pull_getattr('env.T') < max_total_steps:
            t_start = time.time()
            self.pull_state()
            metrics = self.learn(self.buffer_sample(batch_size=batch_size))
            metrics['throughput/learn_loop'] = throughput
            self.push_state()
            self.push_metrics(metrics)
            throughput = batch_size / (time.time() - t_start)

    def buffer_len(self):
        if self.param_store is None:
            assert self.buffer is not None
            len_ = len(self.buffer)
        elif isinstance(self.param_store, self.__ray.actor.ActorHandle):
            len_ = self.__ray.get(self.param_store.buffer_len.remote())
        else:
            len_ = self.param_store.buffer_len()
        return len_

    def buffer_add(self, transition_batch, Adv=None):
        if self.param_store is None:
            assert self.buffer is not None
            if 'Adv' in inspect.signature(self.buffer.add).parameters:  # duck typing
                self.buffer.add(transition_batch, Adv=Adv)
            else:
                self.buffer.add(transition_batch)
        elif isinstance(self.param_store, self.__ray.actor.ActorHandle):
            self.__ray.get(self.param_store.buffer_add.remote(transition_batch, Adv))
        else:
            self.param_store.buffer_add(transition_batch, Adv)

    def buffer_update(self, transition_batch_idx, Adv):
        if self.param_store is None:
            assert self.buffer is not None
            self.buffer.update(transition_batch_idx, Adv=Adv)
        elif isinstance(self.param_store, self.__ray.actor.ActorHandle):
            self.__ray.get(self.param_store.buffer_update.remote(transition_batch_idx, Adv))
        else:
            self.param_store.buffer_update(transition_batch_idx, Adv)

    def buffer_sample(self, batch_size=32):
        buffer_warmup = max(self.buffer_warmup or 0, batch_size)
        wait_secs = 1 / 1024.
        buffer_len = self.buffer_len()
        while buffer_len < buffer_warmup:
            self.env.logger.debug(
                f"buffer insufficiently populated: {buffer_len}/{buffer_warmup}; "
                f"waiting for {wait_secs}s")
            time.sleep(wait_secs)
            wait_secs = min(30, wait_secs * 2)  # wait at most 30s between tries
            buffer_len = self.buffer_len()

        if self.param_store is None:
            assert self.buffer is not None
            transition_batch = self.buffer.sample(batch_size=batch_size)
        elif isinstance(self.param_store, self.__ray.actor.ActorHandle):
            transition_batch = self.__ray.get(
                self.param_store.buffer_sample.remote(batch_size=batch_size))
        else:
            transition_batch = self.param_store.buffer_sample(batch_size=batch_size)
        assert transition_batch is not None
        return transition_batch

    def pull_state(self):
        assert self.param_store is not None, "cannot call pull_state on param_store itself"
        if isinstance(self.param_store, self.__ray.actor.ActorHandle):
            self.set_state(self.__ray.get(self.param_store.get_state.remote()))
        else:
            self.set_state(self.param_store.get_state())

    def push_state(self):
        assert self.param_store is not None, "cannot call push_state on param_store itself"
        if isinstance(self.param_store, self.__ray.actor.ActorHandle):
            self.__ray.get(self.param_store.set_state.remote(self.get_state()))
        else:
            self.param_store.set_state(self.get_state())

    def pull_metrics(self):
        if self.param_store is None:
            metrics = self.env.get_metrics()
        elif isinstance(self.param_store, self.__ray.actor.ActorHandle):
            metrics = self.__ray.get(self.param_store.pull_metrics.remote()).copy()
        else:
            metrics = self.param_store.pull_metrics()
        return metrics

    def push_metrics(self, metrics):
        if self.param_store is None:
            self.env.record_metrics(metrics)
        elif isinstance(self.param_store, self.__ray.actor.ActorHandle):
            self.__ray.get(self.param_store.push_metrics.remote(metrics))
        else:
            self.param_store.push_metrics(metrics)

    def pull_getattr(self, name, default_value=...):
        if self.param_store is None:
            value = _getattr_recursive(self, name, default_value)
        elif isinstance(self.param_store, self.__ray.actor.ActorHandle):
            value = self.__ray.get(self.param_store.pull_getattr.remote(name, default_value))
        else:
            value = self.param_store.pull_getattr(name, default_value)
        return value

    def push_setattr(self, name, value):
        if self.param_store is None:
            _setattr_recursive(self, name, value)
        elif isinstance(self.param_store, self.__ray.actor.ActorHandle):
            self.__ray.get(self.param_store.push_setattr.remote(name, value))
        else:
            self.param_store.push_setattr(name, value)


# -- some helper functions (boilerplate) --------------------------------------------------------- #


def _check_env(env, name):
    if isinstance(env, gym.Env):
        pass
    elif isinstance(env, str):
        env = gym.make(env)
    elif hasattr(env, '__call__'):
        env = env()
    else:
        raise TypeError(f"env must be a gym.Env, str or callable; got: {type(env)}")

    if getattr(getattr(env, 'spec', None), 'max_episode_steps', None) is None:
        raise ValueError(
            "env.spec.max_episode_steps not set; please register env with "
            "gym.register('Foo-v0', entry_point='foo.Foo', max_episode_steps=...) "
            "or wrap your env with: env = gym.wrappers.TimeLimit(env, max_episode_steps=...)")

    if not isinstance(env, TrainMonitor):
        env = TrainMonitor(env, name=name, log_all_metrics=True)

    return env


def _check_reward_threshold(reward_threshold, env):
    if reward_threshold is None:
        reward_threshold = getattr(getattr(env, 'spec', None), 'reward_threshold', None)
    if reward_threshold is None:
        reward_threshold = float('inf')
    return reward_threshold


def _getattr_recursive(obj, name, default=...):
    if '.' not in name:
        return getattr(obj, name) if default is Ellipsis else getattr(obj, name, default)

    name, subname = name.split('.', 1)
    return _getattr_recursive(getattr(obj, name), subname, default)


def _setattr_recursive(obj, name, value):
    if '.' not in name:
        return setattr(obj, name, value)

    name, subname = name.split('.', 1)
    return _setattr_recursive(getattr(obj, name), subname, value)



from abc import ABC, abstractmethod

import jax
import jax.numpy as jnp
import haiku as hk
import optax
import chex

from .._base.mixins import RandomStateMixin
from ..utils import get_grads_diagnostics, is_policy, is_stochastic, is_qfunction, is_vfunction, jit
from ..value_losses import huber, quantile_huber
from ..regularizers import Regularizer
from ..proba_dists import DiscretizedIntervalDist, EmpiricalQuantileDist


__all__ = (
    'BaseTDLearningV',
    'BaseTDLearningQ',
    'BaseTDLearningQWithTargetPolicy'
)


class BaseTDLearning(ABC, RandomStateMixin):
    def __init__(self, f, f_targ=None, optimizer=None, loss_function=None, policy_regularizer=None):

        self._f = f
        self._f_targ = f if f_targ is None else f_targ
        self.loss_function = huber if loss_function is None else loss_function

        if not isinstance(policy_regularizer, (Regularizer, type(None))):
            raise TypeError(
                f"policy_regularizer must be a Regularizer, got: {type(policy_regularizer)}")
        self.policy_regularizer = policy_regularizer

        # optimizer
        self._optimizer = optax.adam(1e-3) if optimizer is None else optimizer
        self._optimizer_state = self.optimizer.init(self._f.params)

        def apply_grads_func(opt, opt_state, params, grads):
            updates, new_opt_state = opt.update(grads, opt_state, params)
            new_params = optax.apply_updates(params, updates)
            return new_opt_state, new_params

        self._apply_grads_func = jit(apply_grads_func, static_argnums=0)

    @abstractmethod
    def target_func(self, target_params, target_state, rng, transition_batch):
        pass

    @property
    @abstractmethod
    def target_params(self):
        pass

    @property
    @abstractmethod
    def target_function_state(self):
        pass

    def update(self, transition_batch, return_td_error=False):
        r"""

        Update the model parameters (weights) of the underlying function approximator.

        Parameters
        ----------
        transition_batch : TransitionBatch

            A batch of transitions.

        return_td_error : bool, optional

            Whether to return the TD-errors.

        Returns
        -------
        metrics : dict of scalar ndarrays

            The structure of the metrics dict is ``{name: score}``.

        td_error : ndarray, optional

            The non-aggregated TD-errors, :code:`shape == (batch_size,)`. This is only returned if
            we set :code:`return_td_error=True`.

        """
        grads, function_state, metrics, td_error = self.grads_and_metrics(transition_batch)
        if any(jnp.any(jnp.isnan(g)) for g in jax.tree_leaves(grads)):
            raise RuntimeError(f"found nan's in grads: {grads}")
        self.apply_grads(grads, function_state)
        return (metrics, td_error) if return_td_error else metrics

    def apply_grads(self, grads, function_state):
        r"""

        Update the model parameters (weights) of the underlying function approximator given
        pre-computed gradients.

        This method is useful in situations in which computation of the gradients is deligated to a
        separate (remote) process.

        Parameters
        ----------
        grads : pytree with ndarray leaves

            A batch of gradients, generated by the :attr:`grads` method.

        function_state : pytree

            The internal state of the forward-pass function. See :attr:`Q.function_state
            <coax.Q.function_state>` and :func:`haiku.transform_with_state` for more details.

        """
        self._f.function_state = function_state
        self.optimizer_state, self._f.params = \
            self._apply_grads_func(self.optimizer, self.optimizer_state, self._f.params, grads)

    def grads_and_metrics(self, transition_batch):
        r"""

        Compute the gradients associated with a batch of transitions.

        Parameters
        ----------
        transition_batch : TransitionBatch

            A batch of transitions.

        Returns
        -------
        grads : pytree with ndarray leaves

            A batch of gradients.

        function_state : pytree

            The internal state of the forward-pass function. See :attr:`Q.function_state
            <coax.Q.function_state>` and :func:`haiku.transform_with_state` for more details.

        metrics : dict of scalar ndarrays

            The structure of the metrics dict is ``{name: score}``.

        td_error : ndarray

            The non-aggregated TD-errors, :code:`shape == (batch_size,)`.

        """
        return self._grads_and_metrics_func(
            self._f.params, self.target_params, self._f.function_state, self.target_function_state,
            self._f.rng, transition_batch)

    def td_error(self, transition_batch):
        r"""

        Compute the TD-errors associated with a batch of transitions. We define the TD-error as the
        negative gradient of the :attr:`loss_function` with respect to the predicted value:

        .. math::

            \text{td_error}_i\ =\ -\frac{\partial L(y, \hat{y})}{\partial \hat{y}_i}

        Note that this reduces to the ordinary definition :math:`\text{td_error}=y-\hat{y}` when we
        use the :func:`coax.value_losses.mse` loss funtion.

        Parameters
        ----------
        transition_batch : TransitionBatch

            A batch of transitions.

        Returns
        -------
        td_errors : ndarray, shape: [batch_size]

            A batch of TD-errors.

        """
        return self._td_error_func(
            self._f.params, self.target_params, self._f.function_state, self.target_function_state,
            self._f.rng, transition_batch)

    @property
    def optimizer(self):
        return self._optimizer

    @optimizer.setter
    def optimizer(self, new_optimizer):
        new_optimizer_state_structure = jax.tree_structure(new_optimizer.init(self._f.params))
        if new_optimizer_state_structure != jax.tree_structure(self.optimizer_state):
            raise AttributeError("cannot set optimizer attr: mismatch in optimizer_state structure")
        self._optimizer = new_optimizer

    @property
    def optimizer_state(self):
        return self._optimizer_state

    @optimizer_state.setter
    def optimizer_state(self, new_optimizer_state):
        if jax.tree_structure(new_optimizer_state) != jax.tree_structure(self.optimizer_state):
            raise AttributeError("cannot set optimizer_state attr: mismatch in tree structure")
        self._optimizer_state = new_optimizer_state


class BaseTDLearningV(BaseTDLearning):
    def __init__(self, v, v_targ=None, optimizer=None, loss_function=None, policy_regularizer=None):

        if not is_vfunction(v):
            raise TypeError(f"v must be a v-function, got: {type(v)}")
        if not (v_targ is None or is_vfunction(v_targ)):
            raise TypeError(f"v_targ must be a v-function or None, got: {type(v_targ)}")

        super().__init__(
            f=v,
            f_targ=v_targ,
            optimizer=optimizer,
            loss_function=loss_function,
            policy_regularizer=policy_regularizer)

        def loss_func(params, target_params, state, target_state, rng, transition_batch):
            """

            In this function we tie together all the pieces, which is why it's a bit long.

            The main structure to watch for is calls to self.target_func(...), which is defined
            downstream. All other code is essentially boilerplate to tie this target to the
            predictions, i.e. to construct a feedback signal for training.

            One of the things we might change here is not to handle both the stochastic and
            deterministic cases in the same function.

            -kris

            """
            rngs = hk.PRNGSequence(rng)
            S = self.v.observation_preprocessor(next(rngs), transition_batch.S)
            W = jnp.clip(transition_batch.W, 0.1, 10.)  # clip importance weights to reduce variance

            metrics = {}

            # regularization term
            if self.policy_regularizer is None:
                regularizer = 0.
            else:
                regularizer, regularizer_metrics = self.policy_regularizer.batch_eval(
                    target_params['reg'], target_params['reg_hparams'], target_state['reg'],
                    next(rngs), transition_batch)
                metrics.update({f'{self.__class__.__name__}/{k}': v for k,
                               v in regularizer_metrics.items()})

            if is_stochastic(self.v):
                dist_params, state_new = self.v.function(params, state, next(rngs), S, True)
                dist_params_target = \
                    self.target_func(target_params, target_state, rng, transition_batch)

                if self.policy_regularizer is not None:
                    dist_params_target = self.v.proba_dist.affine_transform(
                        dist_params_target, 1., -regularizer, self.v.value_transform)

                if isinstance(self.v.proba_dist, DiscretizedIntervalDist):
                    loss = jnp.mean(self.v.proba_dist.cross_entropy(dist_params_target,
                                                                    dist_params))
                elif isinstance(self.v.proba_dist, EmpiricalQuantileDist):
                    loss = quantile_huber(dist_params_target['values'],
                                          dist_params['values'],
                                          dist_params['quantile_fractions'], W)

                # the rest here is only needed for metrics dict
                V = self.v.proba_dist.mean(dist_params)
                V = self.v.proba_dist.postprocess_variate(next(rngs), V, batch_mode=True)
                G = self.v.proba_dist.mean(dist_params_target)
                G = self.v.proba_dist.postprocess_variate(next(rngs), G, batch_mode=True)
                dist_params_v_targ, _ = self.v.function(
                    target_params['v_targ'], target_state['v_targ'], next(rngs), S, False)
                V_targ = self.v.proba_dist.mean(dist_params_v_targ)
                V_targ = self.v.proba_dist.postprocess_variate(next(rngs), V_targ, batch_mode=True)

            else:
                V, state_new = self.v.function(params, state, next(rngs), S, True)
                G = self.target_func(target_params, target_state, next(rngs), transition_batch)
                # flip sign (typical example: regularizer = -beta * entropy)
                G -= regularizer
                loss = self.loss_function(G, V, W)

                # only needed for metrics dict
                V_targ, _ = self.v.function(
                    target_params['v_targ'], target_state['v_targ'], next(rngs), S, False)

            chex.assert_equal_shape([G, V, V_targ, W])
            chex.assert_rank([G, V, V_targ, W], 1)
            dLoss_dV = jax.grad(self.loss_function, argnums=1)
            td_error = -V.shape[0] * dLoss_dV(G, V)  # e.g. (G - V) if loss function is MSE
            chex.assert_equal_shape([td_error, W])
            metrics.update({
                f'{self.__class__.__name__}/loss': loss,
                f'{self.__class__.__name__}/td_error': jnp.mean(W * td_error),
                f'{self.__class__.__name__}/td_error_targ': jnp.mean(-dLoss_dV(V, V_targ, W)),
            })
            return loss, (td_error, state_new, metrics)

        def grads_and_metrics_func(
                params, target_params, state, target_state, rng, transition_batch):

            rngs = hk.PRNGSequence(rng)
            grads, (td_error, state_new, metrics) = jax.grad(loss_func, has_aux=True)(
                params, target_params, state, target_state, next(rngs), transition_batch)

            # add some diagnostics about the gradients
            metrics.update(get_grads_diagnostics(grads, f'{self.__class__.__name__}/grads_'))

            return grads, state_new, metrics, td_error

        def td_error_func(params, target_params, state, target_state, rng, transition_batch):
            loss, (td_error, state_new, metrics) =\
                loss_func(params, target_params, state, target_state, rng, transition_batch)
            return td_error

        self._grads_and_metrics_func = jit(grads_and_metrics_func)
        self._td_error_func = jit(td_error_func)

    @property
    def v(self):
        return self._f

    @property
    def v_targ(self):
        return self._f_targ

    @property
    def target_params(self):
        return hk.data_structures.to_immutable_dict({
            'v': self.v.params,
            'v_targ': self.v_targ.params,
            'reg': getattr(getattr(self.policy_regularizer, 'f', None), 'params', None),
            'reg_hparams': getattr(self.policy_regularizer, 'hyperparams', None)})

    @property
    def target_function_state(self):
        return hk.data_structures.to_immutable_dict({
            'v': self.v.function_state,
            'v_targ': self.v_targ.function_state,
            'reg': getattr(getattr(self.policy_regularizer, 'f', None), 'function_state', None)})

    def _get_target_dist_params(self, params, state, rng, transition_batch):
        r"""

        This method applies techniques from the Distributionel RL paper (arxiv:1707.06887) to
        update StochasticQ / StochasticV.

        """
        rngs = hk.PRNGSequence(rng)
        S_next = self.v_targ.observation_preprocessor(next(rngs), transition_batch.S_next)
        scale, shift = transition_batch.In, transition_batch.Rn  # defines affine transformation
        dist_params_next, _ = self.v_targ.function(params, state, next(rngs), S_next, False)
        dist_params_target = self.v_targ.proba_dist.affine_transform(
            dist_params_next, scale, shift, self.v_targ.value_transform)
        return dist_params_target


class BaseTDLearningQ(BaseTDLearning):
    def __init__(self, q, q_targ=None, optimizer=None, loss_function=None, policy_regularizer=None):

        if not is_qfunction(q):
            raise TypeError(f"q must be a q-function, got: {type(q)}")
        if not (q_targ is None or isinstance(q_targ, (list, tuple)) or is_qfunction(q_targ)):
            raise TypeError(f"q_targ must be a q-function or None, got: {type(q_targ)}")

        super().__init__(
            f=q,
            f_targ=q_targ,
            optimizer=optimizer,
            loss_function=loss_function,
            policy_regularizer=policy_regularizer)

        def loss_func(params, target_params, state, target_state, rng, transition_batch):
            """

            In this function we tie together all the pieces, which is why it's a bit long.

            The main structure to watch for is calls to self.target_func(...), which is defined
            downstream. All other code is essentially boilerplate to tie this target to the
            predictions, i.e. to construct a feedback signal for training.

            One of the things we might change here is not to handle both the stochastic and
            deterministic cases in the same function.

            -kris

            """
            rngs = hk.PRNGSequence(rng)
            S = self.q.observation_preprocessor(next(rngs), transition_batch.S)
            A = self.q.action_preprocessor(next(rngs), transition_batch.A)
            W = jnp.clip(transition_batch.W, 0.1, 10.)  # clip importance weights to reduce variance

            metrics = {}

            # regularization term
            if self.policy_regularizer is None:
                regularizer = 0.
            else:
                regularizer, regularizer_metrics = self.policy_regularizer.batch_eval(
                    target_params['reg'], target_params['reg_hparams'], target_state['reg'],
                    next(rngs), transition_batch)
                metrics.update({f'{self.__class__.__name__}/{k}': v for k,
                               v in regularizer_metrics.items()})

            if is_stochastic(self.q):
                dist_params, state_new = \
                    self.q.function_type1(params, state, next(rngs), S, A, True)
                dist_params_target = \
                    self.target_func(target_params, target_state, rng, transition_batch)

                if self.policy_regularizer is not None:
                    dist_params_target = self.q.proba_dist.affine_transform(
                        dist_params_target, 1., -regularizer, self.q.value_transform)

                if isinstance(self.q.proba_dist, DiscretizedIntervalDist):
                    loss = jnp.mean(self.q.proba_dist.cross_entropy(dist_params_target,
                                                                    dist_params))
                elif isinstance(self.q.proba_dist, EmpiricalQuantileDist):
                    loss = quantile_huber(dist_params_target['values'],
                                          dist_params['values'],
                                          dist_params['quantile_fractions'], W)

                # the rest here is only needed for metrics dict
                Q = self.q.proba_dist.mean(dist_params)
                Q = self.q.proba_dist.postprocess_variate(next(rngs), Q, batch_mode=True)
                G = self.q.proba_dist.mean(dist_params_target)
                G = self.q.proba_dist.postprocess_variate(next(rngs), G, batch_mode=True)
                dist_params_q_targ, _ = self.q.function_type1(
                    target_params['q_targ'], target_state['q_targ'], next(rngs), S, A, False)
                Q_targ = self.q.proba_dist.mean(dist_params_q_targ)
                Q_targ = self.q.proba_dist.postprocess_variate(next(rngs), Q_targ, batch_mode=True)

            else:
                Q, state_new = self.q.function_type1(params, state, next(rngs), S, A, True)
                G = self.target_func(target_params, target_state, next(rngs), transition_batch)
                # flip sign (typical example: regularizer = -beta * entropy)
                G -= regularizer
                loss = self.loss_function(G, Q, W)

                # only needed for metrics dict
                Q_targ, _ = self.q.function_type1(
                    target_params['q_targ'], target_state['q_targ'], next(rngs), S, A, False)

            chex.assert_equal_shape([G, Q, Q_targ, W])
            chex.assert_rank([G, Q, Q_targ, W], 1)
            dLoss_dQ = jax.grad(self.loss_function, argnums=1)
            td_error = -Q.shape[0] * dLoss_dQ(G, Q)  # e.g. (G - Q) if loss function is MSE
            chex.assert_equal_shape([td_error, W])
            metrics.update({
                f'{self.__class__.__name__}/loss': loss,
                f'{self.__class__.__name__}/td_error': jnp.mean(W * td_error),
                f'{self.__class__.__name__}/td_error_targ': jnp.mean(-dLoss_dQ(Q, Q_targ, W)),
            })
            return loss, (td_error, state_new, metrics)

        def grads_and_metrics_func(
                params, target_params, state, target_state, rng, transition_batch):

            rngs = hk.PRNGSequence(rng)
            grads, (td_error, state_new, metrics) = jax.grad(loss_func, has_aux=True)(
                params, target_params, state, target_state, next(rngs), transition_batch)

            # add some diagnostics about the gradients
            metrics.update(get_grads_diagnostics(grads, f'{self.__class__.__name__}/grads_'))

            return grads, state_new, metrics, td_error

        def td_error_func(params, target_params, state, target_state, rng, transition_batch):
            loss, (td_error, state_new, metrics) =\
                loss_func(params, target_params, state, target_state, rng, transition_batch)
            return td_error

        self._grads_and_metrics_func = jit(grads_and_metrics_func)
        self._td_error_func = jit(td_error_func)

    @property
    def q(self):
        return self._f

    @property
    def q_targ(self):
        return self._f_targ

    @property
    def target_params(self):
        return hk.data_structures.to_immutable_dict({
            'q': self.q.params,
            'q_targ': self.q_targ.params,
            'reg': getattr(getattr(self.policy_regularizer, 'f', None), 'params', None),
            'reg_hparams': getattr(self.policy_regularizer, 'hyperparams', None)})

    @property
    def target_function_state(self):
        return hk.data_structures.to_immutable_dict({
            'q': self.q.function_state,
            'q_targ': self.q_targ.function_state,
            'reg': getattr(getattr(self.policy_regularizer, 'f', None), 'function_state', None)})

    def _get_target_dist_params(self, params, state, rng, transition_batch, A_next):
        r"""

        This method applies techniques from the Distributionel RL paper (arxiv:1707.06887) to
        update StochasticQ / StochasticV.

        """
        rngs = hk.PRNGSequence(rng)
        S_next = self.q_targ.observation_preprocessor(next(rngs), transition_batch.S_next)
        scale, shift = transition_batch.In, transition_batch.Rn  # defines affine transformation
        dist_params_next, _ = self.q_targ.function_type1(
            params, state, next(rngs), S_next, A_next, False)
        dist_params_target = self.q_targ.proba_dist.affine_transform(
            dist_params_next, scale, shift, self.q_targ.value_transform)
        return dist_params_target


class BaseTDLearningQWithTargetPolicy(BaseTDLearningQ):
    def __init__(
            self, q, pi_targ, q_targ=None, optimizer=None,
            loss_function=None, policy_regularizer=None):

        if pi_targ is not None and not is_policy(pi_targ):
            raise TypeError(f"pi_targ must be a Policy, got: {type(pi_targ)}")

        self.pi_targ = pi_targ
        super().__init__(
            q=q,
            q_targ=q_targ,
            optimizer=optimizer,
            loss_function=loss_function,
            policy_regularizer=policy_regularizer)

    @property
    def target_params(self):
        return hk.data_structures.to_immutable_dict({
            'q': self.q.params,
            'q_targ': self.q_targ.params,
            'pi_targ': getattr(self.pi_targ, 'params', None),
            'reg': getattr(getattr(self.policy_regularizer, 'f', None), 'params', None),
            'reg_hparams': getattr(self.policy_regularizer, 'hyperparams', None)})

    @property
    def target_function_state(self):
        return hk.data_structures.to_immutable_dict({
            'q': self.q.function_state,
            'q_targ': self.q_targ.function_state,
            'pi_targ': getattr(self.pi_targ, 'function_state', None),
            'reg':
                getattr(getattr(self.policy_regularizer, 'f', None), 'function_state', None)})



from gym.spaces import Box

from ..utils import default_preprocessor
from ..proba_dists import DiscretizedIntervalDist, EmpiricalQuantileDist
from ..value_transforms import ValueTransform
from .base_stochastic_func_type1 import BaseStochasticFuncType1


__all__ = (
    'StochasticQ',
)


class StochasticQ(BaseStochasticFuncType1):
    r"""

    A q-function :math:`q(s,a)`, represented by a stochastic function
    :math:`\mathbb{P}_\theta(G_t|S_t=s,A_t=a)`.

    Parameters
    ----------
    func : function

        A Haiku-style function that specifies the forward pass.

    env : gym.Env

        The gym-style environment. This is used to validate the input/output structure of ``func``.

    value_range : tuple of floats, optional

        A pair of floats :code:`(min_value, max_value)`. If no :code:`value_range` is given,
        :code:`num_bins` is the number of bins of the quantile function as in
        `IQN <https://arxiv.org/abs/1806.06923>`_ or `QR-DQN <https://arxiv.org/abs/1710.10044>`_.

    num_bins : int, optional

        If :code:`value_range` is given: The space of rewards is discretized in :code:`num_bins`
        equal sized bins. We use the default setting of 51 as suggested in the
        `Distributional RL <https://arxiv.org/abs/1707.06887>`_ paper.

        Else: The number of fractions of the quantile function of the rewards is defined by
        :code:`num_bins` as in `IQN <https://arxiv.org/abs/1806.06923>`_ or
        `QR-DQN <https://arxiv.org/abs/1710.10044>`_.

    observation_preprocessor : function, optional

        Turns a single observation into a batch of observations in a form that is convenient for
        feeding into :code:`func`. If left unspecified, this defaults to
        :func:`default_preprocessor(env.observation_space) <coax.utils.default_preprocessor>`.

    action_preprocessor : function, optional

        Turns a single action into a batch of actions in a form that is convenient for feeding into
        :code:`func`. If left unspecified, this defaults
        :func:`default_preprocessor(env.action_space) <coax.utils.default_preprocessor>`.

    value_transform : ValueTransform or pair of funcs, optional

        If provided, the target for the underlying function approximator is transformed:

        .. math::

            \tilde{G}_t\ =\ f(G_t)

        This means that calling the function involves undoing this transformation using its inverse
        :math:`f^{-1}`. The functions :math:`f` and :math:`f^{-1}` are given by
        ``value_transform.transform_func`` and ``value_transform.inverse_func``, respectively. Note
        that a ValueTransform is just a glorified pair of functions, i.e. passing
        ``value_transform=(func, inverse_func)`` works just as well.

    random_seed : int, optional

        Seed for pseudo-random number generators.

    """

    def __init__(
            self, func, env, value_range=None, num_bins=51,
            observation_preprocessor=None, action_preprocessor=None, value_transform=None,
            random_seed=None):

        self.value_transform = value_transform
        proba_dist = self._get_proba_dist(value_transform, num_bins, value_range)

        # set defaults
        if observation_preprocessor is None:
            observation_preprocessor = default_preprocessor(env.observation_space)
        if action_preprocessor is None:
            action_preprocessor = default_preprocessor(env.action_space)
        if self.value_transform is None:
            self.value_transform = ValueTransform(lambda x: x, lambda x: x)
        if not isinstance(self.value_transform, ValueTransform):
            self.value_transform = ValueTransform(*value_transform)

        super().__init__(
            func=func,
            observation_space=env.observation_space,
            action_space=env.action_space,
            observation_preprocessor=observation_preprocessor,
            action_preprocessor=action_preprocessor,
            proba_dist=proba_dist,
            random_seed=random_seed)

    @property
    def num_bins(self):
        return self.proba_dist.space.n

    @classmethod
    def example_data(
            cls, env, value_range, num_bins=51,
            observation_preprocessor=None, action_preprocessor=None, value_transform=None,
            batch_size=1, random_seed=None):

        value_range = cls._check_value_range(value_range)
        proba_dist = cls._get_proba_dist(value_transform, num_bins, value_range)

        if observation_preprocessor is None:
            observation_preprocessor = default_preprocessor(env.observation_space)
        if action_preprocessor is None:
            action_preprocessor = default_preprocessor(env.action_space)

        return super().example_data(
            env=env,
            observation_preprocessor=observation_preprocessor,
            action_preprocessor=action_preprocessor,
            proba_dist=proba_dist,
            batch_size=batch_size,
            random_seed=random_seed)

    def __call__(self, s, a=None, return_logp=False):
        r"""

        Sample a value.

        Parameters
        ----------
        s : state observation

            A single state observation :math:`s`.

        a : action, optional

            A single action :math:`a`. This is *required* if the actions space is non-discrete.

        return_logp : bool, optional

            Whether to return the log-propensity associated with the sampled output value.

        Returns
        -------
        value : float or list thereof

            Depending on whether :code:`a` is provided, this either returns a single value or a list
            of :math:`n` values, one for each discrete action.

        logp : non-positive float or list thereof, optional

            The log-propensity associated with the sampled output value. This is only returned if we
            set ``return_logp=True``. Depending on whether :code:`a` is provided, this is either a
            single float or a list of :math:`n` floats, one for each discrete action.

        """
        return super().__call__(s, a=a, return_logp=return_logp)

    def mean(self, s, a=None):
        r"""

        Get the mean value.

        Parameters
        ----------
        s : state observation

            A single state observation :math:`s`.

        a : action, optional

            A single action :math:`a`. This is *required* if the actions space is non-discrete.

        Returns
        -------
        value : float or list thereof

            Depending on whether :code:`a` is provided, this either returns a single value or a list
            of :math:`n` values, one for each discrete action.

        """
        return super().mean(s, a=a)

    def mode(self, s, a=None):
        r"""

        Get the most probable value.

        Parameters
        ----------
        s : state observation

            A single state observation :math:`s`.

        a : action, optional

            A single action :math:`a`. This is *required* if the actions space is non-discrete.

        Returns
        -------
        value : float or list thereof

            Depending on whether :code:`a` is provided, this either returns a single value or a list
            of :math:`n` values, one for each discrete action.

        """
        return super().mode(s, a=a)

    def dist_params(self, s, a=None):
        r"""

        Get the parameters of the underlying (conditional) probability distribution.

        Parameters
        ----------
        s : state observation

            A single state observation :math:`s`.

        a : action, optional

            A single action :math:`a`. This is *required* if the actions space is non-discrete.

        Returns
        -------
        dist_params : dict or list of dicts

            Depending on whether :code:`a` is provided, this either returns a single dist-params
            dict or a list of :math:`n` such dicts, one for each discrete action.

        """
        return super().dist_params(s, a=a)

    @staticmethod
    def _get_proba_dist(value_transform, num_bins, value_range):
        if value_range is not None:
            if value_transform is not None:
                f, _ = value_transform
                value_range = f(value_range[0]), f(value_range[1])
            reward_space = Box(*value_range, shape=())
            return DiscretizedIntervalDist(reward_space, num_bins)
        else:
            return EmpiricalQuantileDist(num_quantiles=num_bins)

    @staticmethod
    def _check_value_range(value_range):
        if not (isinstance(value_range, (tuple, list))
                and len(value_range) == 2
                and isinstance(value_range[0], (int, float))
                and isinstance(value_range[1], (int, float))
                and value_range[0] < value_range[1]):
            raise TypeError("value_range is not a valid pair tuple of floats: (low, high)")
        return float(value_range[0]), float(value_range[1])



from functools import partial

import jax
import jax.numpy as jnp
import numpy as onp

from .._base.mixins import CopyMixin
from ..utils import pretty_repr


__all__ = (
    'TransitionBatch',
)


class TransitionBatch(CopyMixin):
    r"""

    A container object for a batch of MDP transitions.

    Parameters
    ----------

    S : pytree with ndarray leaves

        A batch of state observations :math:`S_t`.

    A : ndarray

        A batch of actions :math:`A_t`.

    logP : ndarray

        A batch of log-propensities :math:`\log\pi(A_t|S_t)`.

    Rn : ndarray

        A batch of partial (:math:`\gamma`-discounted) returns. For instance,
        in :math:`n`-step bootstrapping these are given by:

        .. math::

            R^{(n)}_t\ &=\ \sum_{k=0}^{n-1}\gamma^kR_{t+k} \\

        In other words, it's the part of the :math:`n`-step return *without*
        the bootstrapping term.

    In : ndarray

        A batch of bootstrap factors. For instance, in :math:`n`-step
        bootstrapping these are given by :math:`I^{(n)}_t=\gamma^n` when
        bootstrapping and :math:`I^{(n)}_t=0` otherwise. Bootstrap factors are
        used in constructing the :math:`n`-step bootstrapped target:

        .. math::

            G^{(n)}_t\ =\ R^{(n)}_t + I^{(n)}_t\,Q(S_{t+1}, A_{t+1})

    S_next : pytree with ndarray leaves

        A batch of next-state observations :math:`S_{t+n}`. This is typically
        used to contruct the TD target in :math:`n`-step bootstrapping.

    A_next : ndarray, optional

        A batch of next-actions :math:`A_{t+n}`. This is typically used to
        contruct the TD target in :math:`n`-step bootstrapping when using SARSA
        updates.

    logP_next : ndarray, optional

        A batch of log-propensities :math:`\log\pi(A_{t+n}|S_{t+n})`.

    W : ndarray, optional

        A batch of importance weights associated with the sampling procedure that generated each
        transition. For example, we need these values when we sample transitions from a
        :class:`PrioritizedReplayBuffer <coax.experience_replay.PrioritizedReplayBuffer>`.

    """
    __slots__ = ('S', 'A', 'logP', 'Rn', 'In', 'S_next',
                 'A_next', 'logP_next', 'W', 'idx', 'extra_info')

    def __init__(self, S, A, logP, Rn, In, S_next, A_next=None, logP_next=None, W=None, idx=None,
                 extra_info=None):

        self.S = S
        self.A = A
        self.logP = logP
        self.Rn = Rn
        self.In = In
        self.S_next = S_next
        self.A_next = A_next
        self.logP_next = logP_next
        self.W = onp.ones_like(Rn) if W is None else W
        self.idx = onp.arange(Rn.shape[0], dtype='int32') if idx is None else idx
        self.extra_info = extra_info

    @classmethod
    def from_single(
            cls, s, a, logp, r, done, gamma,
            s_next=None, a_next=None, logp_next=None, w=1, idx=None, extra_info=None):
        r"""

        Create a TransitionBatch (with batch_size=1) from a single transition.

        Attributes
        ----------
        s : state observation

            A single state observation :math:`S_t`.

        a : action

            A single action :math:`A_t`.

        logp : non-positive float

            The log-propensity :math:`\log\pi(A_t|S_t)`.

        r : float or array of floats

            A single reward :math:`R_t`.

        done : bool

            Whether the episode has finished.

        info : dict or None

            Some additional info about the current time step.

        s_next : state observation

            A single next-state observation :math:`S_{t+1}`.

        a_next : action

            A single next-action :math:`A_{t+1}`.

        logp_next : non-positive float

            The log-propensity :math:`\log\pi(A_{t+1}|S_{t+1})`.

        w : positive float, optional

            The importance weight associated with the sampling procedure that generated this
            transition.

        idx : int, optional

            The identifier of this particular transition.

        """

        # check types
        array = (int, float, onp.ndarray, jnp.ndarray)
        if not (isinstance(logp, array) and onp.all(logp <= 0)):
            raise TypeError(f"logp must be non-positive float(s), got: {logp}")
        if not isinstance(r, array):
            raise TypeError(f"r must be a scalar or an array, got: {r}")
        if not isinstance(done, bool):
            raise TypeError(f"done must be a bool, got: {done}")
        if not (isinstance(gamma, (float, int)) and 0 <= gamma <= 1):
            raise TypeError(f"gamma must be a float in the unit interval [0, 1], got: {gamma}")
        if not (logp_next is None or (isinstance(logp_next, array) and onp.all(logp_next <= 0))):
            raise TypeError(f"logp_next must be None or non-positive float(s), got: {logp_next}")
        if not (isinstance(w, (float, int)) and w > 0):
            raise TypeError(f"w must be a positive float, got: {w}")

        return cls(
            S=_single_to_batch(s),
            A=_single_to_batch(a),
            logP=_single_to_batch(logp),
            Rn=_single_to_batch(r),
            In=_single_to_batch(float(gamma) * (1. - bool(done))),
            S_next=_single_to_batch(s_next) if s_next is not None else None,
            A_next=_single_to_batch(a_next) if a_next is not None else None,
            logP_next=_single_to_batch(logp_next) if logp_next is not None else None,
            W=_single_to_batch(float(w)),
            idx=_single_to_batch(idx) if idx is not None else None,
            extra_info=_single_to_batch(extra_info) if extra_info is not None else None
        )

    @property
    def batch_size(self):
        return onp.shape(self.Rn)[0]

    def to_singles(self):
        r"""

        Get an iterator of single transitions.

        Returns
        -------
        transition_batches : iterator of TransitionBatch

            An iterator of :class:`TransitionBatch <coax.reward_tracing.TransitionBatch>` objects
            with ``batch_size=1``.

            **Note:** The iterator walks through the individual transitions *in reverse order*.

        """
        if self.batch_size == 1:
            yield self
            return  # break out of generator

        def lookup(i, pytree):
            s = slice(i, i + 1)  # ndim-preserving lookup
            return jax.tree_map(lambda leaf: leaf[s], pytree)

        for i in range(self.batch_size):
            yield TransitionBatch(*map(partial(lookup, i), self))

    def items(self):
        for k in self.__slots__:
            yield k, getattr(self, k)

    def _asdict(self):
        return dict(self.items())

    def __repr__(self):
        return pretty_repr(self)

    def __iter__(self):
        return (getattr(self, a) for a in self.__slots__)

    def __getitem__(self, int_or_slice):
        return tuple(self).__getitem__(int_or_slice)

    def __eq__(self, other):
        return (type(self) is type(other)) and all(
            onp.allclose(a, b) if isinstance(a, (onp.ndarray, jnp.ndarray))
            else (a is b if a is None else a == b)
            for a, b in zip(jax.tree_leaves(self), jax.tree_leaves(other)))


def _single_to_batch(pytree):
    # notice that we're pulling eveyrthing out of jax.numpy and into ordinary numpy land
    return jax.tree_map(lambda arr: onp.expand_dims(arr, axis=0), pytree)


jax.tree_util.register_pytree_node(
    TransitionBatch,
    lambda tn: (tuple(tn), None),
    lambda treedef, leaves: TransitionBatch(*leaves))




class CoaxError(Exception):
    pass


class SpaceError(CoaxError):
    pass


class ActionSpaceError(SpaceError):
    pass


class DistributionError(CoaxError):
    pass


class EpisodeDoneError(CoaxError):
    pass


class InconsistentCacheInputError(CoaxError):
    pass


class InsufficientCacheError(CoaxError):
    pass


class LeafNodeError(CoaxError):
    pass


class MissingAdversaryError(CoaxError):
    pass


class MissingModelError(CoaxError):
    pass


class NotLeafNodeError(CoaxError):
    pass


class NumpyArrayCheckError(CoaxError):
    pass


class TensorCheckError(CoaxError):
    pass


class UnavailableActionError(CoaxError):
    pass



from .._base.errors import InsufficientCacheError, EpisodeDoneError
from ._base import BaseRewardTracer
from ._transition import TransitionBatch


__all__ = (
    'MonteCarlo',
)


class MonteCarlo(BaseRewardTracer):
    r"""
    A short-term cache for episodic Monte Carlo sampling.

    Parameters
    ----------
    gamma : float between 0 and 1

        The amount by which to discount future rewards.

    """
    def __init__(self, gamma):
        self.gamma = float(gamma)
        self.reset()

    def reset(self):
        self._list = []
        self._done = False
        self._g = 0  # accumulator for return

    def add(self, s, a, r, done, logp=0.0, w=1.0):
        if self._done and len(self):
            raise EpisodeDoneError(
                "please flush cache (or repeatedly pop) before appending new transitions")

        self._list.append((s, a, r, logp, w))
        self._done = bool(done)
        if self._done:
            self._g = 0.  # init return

    def __len__(self):
        return len(self._list)

    def __bool__(self):
        return bool(len(self)) and self._done

    def pop(self):
        if not self:
            if not len(self):
                raise InsufficientCacheError(
                    "cache needs to receive more transitions before it can be popped from")
            else:
                raise InsufficientCacheError(
                    "cannot pop from cache before before receiving done=True")

        # pop state-action (propensities) pair
        s, a, r, logp, w = self._list.pop()

        # update return
        self._g = r + self.gamma * self._g

        return TransitionBatch.from_single(
            s=s, a=a, logp=logp, r=self._g, done=True, gamma=self.gamma,  # no bootstrapping
            s_next=s, a_next=a, logp_next=logp, w=w)                      # dummy values for *_next



r"""
Model Updaters
==============

.. autosummary::
    :nosignatures:

    coax.model_updaters.ModelUpdater

----

This is a collection of objects that are used to update dynamics models, i.e. transition models and
reward functions.


Object Reference
----------------

.. autoclass:: coax.model_updaters.ModelUpdater

"""

# TODO(krholshe): think of better names for this module and classes

from ._model_updater import ModelUpdater


__all__ = (
    'ModelUpdater',
)



import numpy as onp
import jax


class RandomStateMixin:
    @property
    def random_seed(self):
        return self._random_seed

    @random_seed.setter
    def random_seed(self, new_random_seed):
        if new_random_seed is None:
            new_random_seed = onp.random.randint(2147483647)
        self._random_seed = new_random_seed
        self._random_key = jax.random.PRNGKey(self._random_seed)

    @property
    def rng(self):
        self._random_key, key = jax.random.split(self._random_key)
        return key



import numpy as onp


__all__ = (
    'SegmentTree',
    'SumTree',
    'MinTree',
    'MaxTree',
)


class SegmentTree:
    r"""

    A `segment tree <https://en.wikipedia.org/wiki/Segment_tree>`_ data structure that allows
    for batched updating and batched partial-range (segment) reductions.

    Parameters
    ----------
    capacity : positive int

        Number of values to accommodate.

    reducer : function

        The reducer function: :code:`(float, float) -> float`.

    init_value : float

        The unit element relative to the reducer function. Some typical examples are: 0 if reducer
        is :data:`add <numpy.add>`, 1 for :data:`multiply <numpy.multiply>`, :math:`-\infty` for
        :data:`maximum <numpy.maximum>`, :math:`\infty` for :data:`minimum <numpy.minimum>`.

    Warning
    -------

    The :attr:`values` attribute and square-bracket lookups (:code:`tree[level, index]`) return
    references of the underlying storage array. Therefore, make sure that downstream code doesn't
    update these values in-place, which would corrupt the segment tree structure.

    """
    def __init__(self, capacity, reducer, init_value):
        self.capacity = capacity
        self.reducer = reducer
        self.init_value = float(init_value)
        self._height = int(onp.ceil(onp.log2(capacity))) + 1  # the +1 is for the values themselves
        self._arr = onp.full(shape=(2 ** self.height - 1), fill_value=self.init_value)

    @property
    def height(self):
        r""" The height of the tree :math:`h\sim\log(\text{capacity})`. """
        return self._height

    @property
    def root_value(self):
        r"""

        The aggregated value, equivalent to
        :func:`reduce(reducer, values, init_value) <functools.reduce>`.

        """
        return self._arr[0]

    @property
    def values(self):
        r""" The values stored at the leaves of the tree. """
        start = 2 ** (self.height - 1) - 1
        stop = start + self.capacity
        return self._arr[start:stop]

    def __getitem__(self, lookup):
        if isinstance(lookup, int):
            level_offset, level_size = self._check_level_lookup(lookup)
            return self._arr[level_offset:(level_offset + level_size)]

        if isinstance(lookup, tuple) and len(lookup) == 1:
            level, = lookup
            return self[level]

        if isinstance(lookup, tuple) and len(lookup) == 2:
            level, index = lookup
            return self[level][index]

        raise IndexError(
            "tree lookup must be of the form: tree[level] or tree[level, index], "
            "where 'level' is an int and 'index' is a 1d array lookup")

    def set_values(self, idx, values):
        r"""

        Set or update the :attr:`values`.

        Parameters
        ----------
        idx : 1d array of ints

            The indices of the values to be updated. If you wish to update all values use ellipses
            instead, e.g. :code:`tree.set_values(..., values)`.

        values : 1d array of floats

            The new values.

        """
        idx, level_offset, level_size = self._check_idx(idx)

        # update leaf-node values
        self._arr[level_offset + (idx % level_size)] = values

        for level in range(self.height - 2, -1, -1):
            idx = onp.unique(idx // 2)
            left_child = level_offset + 2 * idx
            right_child = left_child + 1

            level_offset = 2 ** level - 1
            parent = level_offset + idx
            self._arr[parent] = self.reducer(self._arr[left_child], self._arr[right_child])

    def partial_reduce(self, start=0, stop=None):
        r"""

        Reduce values over a partial range of indices. This is an efficient, batched implementation
        of :func:`reduce(reducer, values[state:stop], init_value) <functools.reduce>`.

        Parameters
        ----------
        start : int or array of ints

            The lower bound of the range (inclusive).

        stop : int or array of ints, optional

            The lower bound of the range (exclusive). If left unspecified, this defaults to
            :attr:`height`.

        Returns
        -------
        value : float

            The result of the partial reduction.

        """
        # NOTE: This is an iterative implementation, which is a lot uglier than a recursive one.
        # The reason why we use an iterative approach is that it's easier for batch-processing.

        # i and j are 1d arrays (indices for self._arr)
        i, j = self._check_start_stop_to_i_j(start, stop)

        # trivial case
        done = (i == j)
        if done.all():
            return self._arr[i]

        # left/right accumulators (mask one of them to avoid over-counting if i == j)
        a, b = self._arr[i], onp.where(done, self.init_value, self._arr[j])

        # number of nodes in higher levels
        level_offset = 2 ** (self.height - 1) - 1

        # we start from the leaves and work up towards the root
        for level in range(self.height - 2, -1, -1):

            # get parent indices
            level_offset_parent = 2 ** level - 1
            i_parent = (i - level_offset) // 2 + level_offset_parent
            j_parent = (j - level_offset) // 2 + level_offset_parent

            # stop when we have a shared parent (possibly the root node, but not necessarily)
            done |= (i_parent == j_parent)
            if done.all():
                return self.reducer(a, b)

            # only accumulate right-child value if 'i' was a left child of 'i_parent'
            a = onp.where((i % 2 == 1) & ~done, self.reducer(a, self._arr[i + 1]), a)

            # only accumulate left-child value if 'j' was a right child of 'j_parent'
            b = onp.where((j % 2 == 0) & ~done, self.reducer(b, self._arr[j - 1]), b)

            # prepare for next loop
            i, j, level_offset = i_parent, j_parent, level_offset_parent

        assert False, 'this point should not be reached'

    def __repr__(self):
        s = ""
        for level in range(self.height):
            s += f"\n  level={level} : {repr(self[level])}"
        return f"{type(self).__name__}({s})"

    def _check_level_lookup(self, level):
        if not isinstance(level, int):
            raise IndexError(f"level lookup must be an int, got: {type(level)}")

        if not (-self.height <= level < self.height):
            raise IndexError(f"level index {level} is out of bounds; tree height: {self.height}")

        level %= self.height
        level_offset = 2 ** level - 1
        level_size = min(2 ** level, self.capacity)
        return level_offset, level_size

    def _check_level(self, level):
        if level < -self.height or level >= self.height:
            raise IndexError(f"tree level index {level} out of range; tree height: {self.height}")
        return level % self.height

    def _check_idx(self, idx):
        """ some boiler plate to turn any compatible idx into a 1d integer array """
        level_offset, level_size = self._check_level_lookup(self.height - 1)

        if isinstance(idx, int):
            idx = onp.asarray([idx], dtype='int32')
        if idx is None or idx is Ellipsis:
            idx = onp.arange(level_size, dtype='int32')
        elif isinstance(idx, list) and all(isinstance(x, int) for x in idx):
            idx = onp.asarray(idx, dtype='int32')
        elif (isinstance(idx, onp.ndarray)
                and onp.issubdtype(idx.dtype, onp.integer)
                and idx.ndim <= 1):
            idx = idx.reshape(-1)
        else:
            raise IndexError("idx must be an int or a 1d integer array")

        if not onp.all((idx < level_size) & (idx >= -level_size)):
            raise IndexError("one of more entries in idx are out or range")

        return idx % level_size, level_offset, level_size

    def _check_start_stop_to_i_j(self, start, stop):
        """ some boiler plate to turn (start, stop) into left/right index arrays (i, j) """
        start_orig, stop_orig = start, stop

        # convert 'start' index to 1d array
        if isinstance(start, int):
            start = onp.array([start])
        if not (isinstance(start, onp.ndarray)
                and start.ndim == 1
                and onp.issubdtype(start.dtype, onp.integer)):
            raise TypeError("'start' must be an int or a 1d integer array")

        # convert 'stop' index to 1d array
        if stop is None:
            stop = onp.full_like(start, self.capacity)
        if isinstance(stop, int):
            stop = onp.full_like(start, stop)
        if not (isinstance(stop, onp.ndarray)
                and stop.ndim == 1
                and onp.issubdtype(stop.dtype, onp.integer)):
            raise TypeError("'stop' must be an int or a 1d integer array")

        # ensure that 'start' is the same size as 'stop'
        if start.size == 1 and stop.size > 1:
            start = onp.full_like(stop, start[0])

        # check compatible shapes
        if start.shape != stop.shape:
            raise ValueError(
                f"shapes must be equal, got: start.shape: {start.shape}, stop.shape: {stop.shape}")

        # convert to (i, j), where j is the *inclusive* version of 'stop' (which is exclusive)
        level_offset, level_size = self._check_level_lookup(self.height - 1)
        i = level_offset + start % level_size
        j = level_offset + (stop - 1) % level_size

        # check consistency of ranges
        if not onp.all((i >= level_offset) & (j < level_offset + level_size) & (i <= j)):
            raise IndexError(
                f"inconsistent ranges detected from (start, stop) = ({start_orig}, {stop_orig})")

        return i, j


class SumTree(SegmentTree):
    r"""

    A sum-tree data structure that allows for batched updating and batched weighted sampling.

    Both update and sampling operations have a time complexity of :math:`\mathcal{O}(\log N)` and a
    memory footprint of :math:`\mathcal{O}(N)`, where :math:`N` is the length of the underlying
    :attr:`values`.

    Parameters
    ----------
    capacity : positive int

        Number of values to accommodate.

    reducer : function

        The reducer function: :code:`(float, float) -> float`.

    init_value : float

        The unit element relative to the reducer function. Some typical examples are: 0 if
        reducer is :func:`operator.add`, 1 for :func:`operator.mul`, :math:`-\infty` for
        :func:`max`, :math:`\infty` for :func:`min`.

    """
    def __init__(self, capacity, random_seed=None):
        super().__init__(capacity=capacity, reducer=onp.add, init_value=0)
        self.random_seed = random_seed

    @property
    def random_seed(self):
        return self._random_seed

    @random_seed.setter
    def random_seed(self, new_random_seed):
        self._rnd = onp.random.RandomState(new_random_seed)
        self._random_seed = new_random_seed

    def sample(self, n):
        r"""

        Sample array indices using weighted sampling, where the sample weights are proprotional to
        the values stored in :attr:`values`.

        Parameters
        ----------
        n : positive int

            The number of samples to return.

        Returns
        -------
        idx : array of ints

            The sampled indices, shape: (n,)

        Warning
        -------

        This method presumes (but doesn't check) that all :attr:`values` stored in the tree are
        non-negative.

        """
        if not (isinstance(n, int) and n > 0):
            raise TypeError("n must be a positive integer")

        return self.inverse_cdf(self._rnd.rand(n))

    def inverse_cdf(self, u):
        r"""

        Inverse of the cumulative distribution function (CDF) of the categorical distribution
        :math:`\text{Cat}(p)`, where :math:`p` are the normalized values :math:`p_i=`
        :attr:`values[i] / sum(values) <values>`.

        This function provides the machinery for the :attr:`sample` method.

        Parameters
        ----------
        u : float or 1d array of floats

            One of more numbers :math:`u\in[0,1]`. These are typically sampled from
            :math:`\text{Unif([0, 1])}`.

        Returns
        -------
        idx : array of ints

            The indices associated with :math:`u`, shape: (n,)

        Warning
        -------

        This method presumes (but doesn't check) that all :attr:`values` stored in the tree are
        non-negative.

        """
        # NOTE: This is an iterative implementation, which is a lot uglier than a recursive one.
        # The reason why we use an iterative approach is that it's easier for batch-processing.
        if self.root_value <= 0:
            raise RuntimeError("the root_value must be positive")

        # init (will be updated in loop)
        u, isscalar = self._check_u(u)
        values = u * self.root_value
        idx = onp.zeros_like(values, dtype='int32')  # this is ultimately what we'll returned
        level_offset_parent = 0                      # number of nodes in levels above parent

        # iterate down, from the root to the leaves
        for level in range(1, self.height):

            # get child indices
            level_offset = 2 ** level - 1
            left_child_idx = (idx - level_offset_parent) * 2 + level_offset
            right_child_idx = left_child_idx + 1

            # update (idx, values, level_offset_parent)
            left_child_values = self._arr[left_child_idx]
            pick_left_child = left_child_values > values
            idx = onp.where(pick_left_child, left_child_idx, right_child_idx)
            values = onp.where(pick_left_child, values, values - left_child_values)
            level_offset_parent = level_offset

        idx = idx - level_offset_parent
        return idx[0] if isscalar else idx

    def _check_u(self, u):
        """ some boilerplate to check validity of 'u' array """
        isscalar = False
        if isinstance(u, (float, int)):
            u = onp.array([u], dtype='float')
            isscalar = True
        if isinstance(u, list) and all(isinstance(x, (float, int)) for x in u):
            u = onp.asarray(u, dtype='float')
        if not (isinstance(u, onp.ndarray)
                and u.ndim == 1 and onp.issubdtype(u.dtype, onp.floating)):
            raise TypeError("'u' must be a float or a 1d array of floats")
        if onp.any(u > 1) or onp.any(u < 0):
            raise ValueError("all values in 'u' must lie in the unit interval [0, 1]")
        return u, isscalar


class MinTree(SegmentTree):
    r"""

    A min-tree data structure, which is a :class:`SegmentTree` whose reducer is :data:`minimum
    <numpy.minimum>`.

    Parameters
    ----------
    capacity : positive int

        Number of values to accommodate.

    """
    def __init__(self, capacity):
        super().__init__(capacity=capacity, reducer=onp.minimum, init_value=float('inf'))


class MaxTree(SegmentTree):
    r"""

    A max-tree data structure, which is a :class:`SegmentTree` whose reducer is :data:`maximum
    <numpy.maximum>`.

    Parameters
    ----------
    capacity : positive int

        Number of values to accommodate.

    """
    def __init__(self, capacity):
        super().__init__(capacity=capacity, reducer=onp.maximum, init_value=-float('inf'))



#!/usr/bin/env python3
import re
import os
import setuptools
from collections import namedtuple

PROJECTDIR = os.path.dirname(__file__)
RE_VERSION = re.compile(
    r'^__version__ \= \'(?P<version>(?P<majorminor>\d+\.\d+)\.\d+(?:\w+\d+)?)\'$', re.MULTILINE)
DEV_STATUS = {
    '0.1': 'Development Status :: 1 - Planning',           # v0.1 - skeleton
    '0.2': 'Development Status :: 2 - Pre-Alpha',          # v0.2 - some basic functionality
    '0.3': 'Development Status :: 3 - Alpha',              # v0.3 - most functionality
    '0.4': 'Development Status :: 4 - Beta',               # v0.4 - most functionality + doc
    '1.0': 'Development Status :: 5 - Production/Stable',  # v1.0 - most functionality + doc + test
    '2.0': 'Development Status :: 6 - Mature',             # v2.0 - new functionality?
}

VersionSpec = namedtuple('VersionSpec', 'version majorminor')


def get_install_requires(requirements_txt):
    install_requires = []
    with open(os.path.join(PROJECTDIR, requirements_txt)) as f:
        for line in f:
            line = line.strip()
            if line and not line.startswith('#'):
                install_requires.append(line)
    return install_requires


def get_version_spec():
    with open(os.path.join(PROJECTDIR, 'coax', '__init__.py')) as f:
        version_match = re.search(RE_VERSION, f.read())
    assert version_match is not None, "can't parse __version__ from __init__.py"
    version_spec = VersionSpec(**version_match.groupdict())
    return version_spec


def get_long_description():
    with open(os.path.join(PROJECTDIR, 'README.rst')) as f:
        return f.read()


version_spec = get_version_spec()


# main setup kw args
setup_kwargs = {
    'name': 'coax',
    'version': version_spec.version,
    'description': "Plug-n-play reinforcement learning with OpenAI Gym and JAX",
    'long_description': get_long_description(),
    'author': 'Kristian Holsheimer',
    'author_email': 'kristian.holsheimer@gmail.com',
    'url': 'https://coax.readthedocs.io',
    'license': 'MIT',
    'python_requires': '~=3.6',
    'install_requires': get_install_requires('requirements.txt'),
    'extras_require': {
        'dev': get_install_requires('requirements.dev.txt'),
        'doc': get_install_requires('requirements.doc.txt'),
        'ray': ['ray>1.9.0'],
    },
    'classifiers': [
        DEV_STATUS[version_spec.majorminor],
        'Environment :: GPU :: NVIDIA CUDA',
        'Framework :: Flake8',
        'Framework :: IPython',
        'Framework :: Jupyter',
        'Framework :: Pytest',
        'Framework :: Sphinx',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    'zip_safe': True,
    'packages': setuptools.find_packages(),
}


if __name__ == '__main__':
    setuptools.setup(**setup_kwargs)



import warnings

import jax.numpy as jnp
import haiku as hk
import chex

from ..utils import check_preprocessors, is_qfunction, is_stochastic
from ._base import PolicyObjective


class DeterministicPG(PolicyObjective):
    r"""
    A deterministic policy-gradient objective, a.k.a. DDPG-style objective. See
    :doc:`spinup:algorithms/ddpg` and references therein for more details.

    .. math::

        J(\theta; s,a)\ =\ q_\text{targ}(s, a_\theta(s))

    Here :math:`a_\theta(s)` is the *mode* of the underlying conditional
    probability distribution :math:`\pi_\theta(.|s)`. See e.g. the :attr:`mode`
    method of :class:`coax.proba_dists.NormalDist`. In other words, we evaluate
    the policy according to the current estimate of its best-case performance.

    This objective has the property that it explicitly maximizes the q-value.

    The full policy loss is constructed as:

    .. math::

        \text{loss}(\theta; s,a)\ =\
            -J(\theta; s,a)
            - \beta_\text{ent}\,H[\pi_\theta]
            + \beta_\text{kl-div}\,
                KL[\pi_{\theta_\text{prior}}, \pi_\theta]

    N.B. in order to unclutter the notation we abbreviated
    :math:`\pi(.|s)` by :math:`\pi`.

    Parameters
    ----------
    pi : Policy

        The parametrized policy :math:`\pi_\theta(a|s)`.

    q_targ : Q

        The target state-action value function :math:`q_\text{targ}(s,a)`.

    optimizer : optax optimizer, optional

        An optax-style optimizer. The default optimizer is :func:`optax.adam(1e-3)
        <optax.adam>`.

    regularizer : Regularizer, optional

        A policy regularizer, see :mod:`coax.regularizers`.

    """
    REQUIRES_PROPENSITIES = False

    def __init__(self, pi, q_targ, optimizer=None, regularizer=None):
        if not is_qfunction(q_targ):
            raise TypeError(f"q must be a q-function, got: {type(q_targ)}")
        if q_targ.modeltype != 1:
            raise TypeError("q must be a type-1 q-function")

        super().__init__(pi=pi, optimizer=optimizer, regularizer=regularizer)
        self.q_targ = q_targ

        if not check_preprocessors(
                self.pi.action_space,
                self.q_targ.action_preprocessor,
                self.pi.proba_dist.preprocess_variate):
            warnings.warn(
                "it seems that q_targ.action_preprocessor does not match "
                "pi.proba_dist.preprocess_variate; please instantiate your q-function using "
                "q = coax.Q(..., action_preprocessor=pi.proba_dist.preprocess_variate) to ensure "
                "that the preprocessors match")

    @property
    def hyperparams(self):
        return hk.data_structures.to_immutable_dict({
            'regularizer': getattr(self.regularizer, 'hyperparams', {}),
            'q': {'params': self.q_targ.params, 'function_state': self.q_targ.function_state}})

    def objective_func(self, params, state, hyperparams, rng, transition_batch, Adv):
        rngs = hk.PRNGSequence(rng)

        # get distribution params from function approximator
        S = self.pi.observation_preprocessor(next(rngs), transition_batch.S)
        dist_params, state_new = self.pi.function(params, state, next(rngs), S, True)

        # compute objective: q(s, a_greedy)
        S = self.q_targ.observation_preprocessor(next(rngs), transition_batch.S)
        A = self.pi.proba_dist.mode(dist_params)
        log_pi = self.pi.proba_dist.log_proba(dist_params, A)
        params_q, state_q = hyperparams['q']['params'], hyperparams['q']['function_state']
        if is_stochastic(self.q_targ):
            dist_params_q, _ = self.q_targ.function_type1(params_q, state_q, rng, S, A, True)
            Q = self.q_targ.proba_dist.mean(dist_params_q)
            Q = self.q_targ.proba_dist.postprocess_variate(next(rngs), Q, batch_mode=True)
        else:
            Q, _ = self.q_targ.function_type1(params_q, state_q, next(rngs), S, A, True)

        # clip importance weights to reduce variance
        W = jnp.clip(transition_batch.W, 0.1, 10.)

        # the objective
        chex.assert_equal_shape([W, Q])
        chex.assert_rank([W, Q], 1)
        objective = W * Q

        return jnp.mean(objective), (dist_params, log_pi, state_new)

    def update(self, transition_batch, Adv=None):
        r"""

        Update the model parameters (weights) of the underlying function approximator.

        Parameters
        ----------
        transition_batch : TransitionBatch

            A batch of transitions.

        Adv : ndarray, ignored

            This input is ignored; it is included for consistency with other policy objectives.

        Returns
        -------
        metrics : dict of scalar ndarrays

            The structure of the metrics dict is ``{name: score}``.

        """
        return super().update(transition_batch, None)

    def grads_and_metrics(self, transition_batch, Adv=None):
        r"""

        Compute the gradients associated with a batch of transitions with
        corresponding advantages.

        Parameters
        ----------
        transition_batch : TransitionBatch

            A batch of transitions.

        Adv : ndarray, ignored

            This input is ignored; it is included for consistency with other policy objectives.

        Returns
        -------
        grads : pytree with ndarray leaves

            A batch of gradients.

        function_state : pytree

            The internal state of the forward-pass function. See :attr:`Policy.function_state
            <coax.Policy.function_state>` and :func:`haiku.transform_with_state` for more details.

        metrics : dict of scalar ndarrays

            The structure of the metrics dict is ``{name: score}``.

        """
        return super().grads_and_metrics(transition_batch, None)



import warnings

import jax
import jax.numpy as jnp
import numpy as onp
from gym.spaces import Box

from ..utils import clipped_logit, jit
from ._base import BaseProbaDist


__all__ = (
    'NormalDist',
)


class NormalDist(BaseProbaDist):
    r"""

    A differentiable normal distribution.

    The input ``dist_params`` to each of the functions is expected to be of the form:

    .. code:: python

        dist_params = {'mu': array([...]), 'logvar': array([...])}

    which represent the (conditional) distribution parameters. Here, ``mu`` is the mean :math:`\mu`
    and ``logvar`` is the log-variance :math:`\log(\sigma^2)`.

    Parameters
    ----------
    space : gym.spaces.Box

        The gym-style space that specifies the domain of the distribution.

    clip_box : pair of floats, optional

        The range of values to allow for *clean* (compact) variates. This is mainly to ensure
        reasonable values when one or more dimensions of the Box space have very large ranges, while
        in reality only a small part of that range is occupied.

    clip_reals : pair of floats, optional

        The range of values to allow for *raw* (decompactified) variates, the *reals*, used
        internally. This range is set for numeric stability. Namely, the :attr:`postprocess_variate`
        method compactifies the reals to a closed interval (Box) by applying a logistic sigmoid.
        Setting a finite range for :code:`clip_reals` ensures that the sigmoid doesn't fully
        saturate.

    """
    def __init__(self, space, clip_box=(-256., 256.), clip_reals=(-30., 30.)):
        if not isinstance(space, Box):
            raise TypeError(f"{self.__class__.__name__} can only be defined over Box spaces")

        super().__init__(space)

        self.clip_box = clip_box
        self.clip_reals = clip_reals

        self._low = onp.maximum(onp.expand_dims(self.space.low, axis=0), self.clip_box[0])
        self._high = onp.minimum(onp.expand_dims(self.space.high, axis=0), self.clip_box[1])
        onp.testing.assert_array_less(
            self._low, self._high,
            "Box clipping resulted in inconsistent boundaries: "
            f"low={self._low}, high={self._high}; please specify proper clipping values, "
            "e.g. NormalDist(space, clip_box=(-1000., 1000.))")
        if onp.any(self._low > self.space.low) or onp.any(self._high < self.space.high):
            with onp.printoptions(precision=1):
                warnings.warn(
                    f"one or more dimensions of Box(low={self.space.low}, high={self.space.high}) "
                    f"will be clipped to Box(low={self._low[0]}, high={self._high[0]})")

        log_2pi = onp.asarray(1.8378770664093453)  # abbreviation

        def check_shape(x, name, flatten):
            if not isinstance(x, jnp.ndarray):
                raise TypeError(f"expected an jax.numpy.ndarray, got: {type(x)}")
            if not (x.ndim == len(space.shape) + 1 and x.shape[1:] == space.shape):
                expected = ', '.join(f'{i:d}' for i in space.shape)
                raise ValueError(f"expected {name}.shape: (?, {expected}), got: {x.shape}")
            if flatten:
                x = x.reshape(x.shape[0], -1)  # batch-flatten
            return x

        def sample(dist_params, rng):
            mu = check_shape(dist_params['mu'], name='mu', flatten=True)
            logvar = check_shape(dist_params['logvar'], name='logvar', flatten=True)

            X = mu + jnp.exp(logvar / 2) * jax.random.normal(rng, mu.shape)
            return X.reshape(-1, *self.space.shape)

        def mean(dist_params):
            return check_shape(dist_params['mu'], name='mu', flatten=False)

        def mode(dist_params):
            return mean(dist_params)

        def log_proba(dist_params, X):
            X = check_shape(X, name='X', flatten=True)
            mu = check_shape(dist_params['mu'], name='mu', flatten=True)
            logvar = check_shape(dist_params['logvar'], name='logvar', flatten=True)

            n = logvar.shape[-1]
            logdetvar = jnp.sum(logvar, axis=-1)  # log(det(M)) = tr(log(M))
            quadratic = jnp.einsum('ij,ij->i', jnp.square(X - mu), jnp.exp(-logvar))
            return -0.5 * (n * log_2pi + logdetvar + quadratic)

        def entropy(dist_params):
            logvar = check_shape(dist_params['logvar'], name='logvar', flatten=True)

            assert logvar.ndim == 2  # check if flattened
            logdetvar = jnp.sum(logvar, axis=-1)  # log(det(M)) = tr(log(M))
            n = logvar.shape[-1]
            return 0.5 * (n * log_2pi + logdetvar + n)

        def cross_entropy(dist_params_p, dist_params_q):
            mu1 = check_shape(dist_params_p['mu'], name='mu_p', flatten=True)
            mu2 = check_shape(dist_params_q['mu'], name='mu_q', flatten=True)
            logvar1 = check_shape(dist_params_p['logvar'], name='logvar_p', flatten=True)
            logvar2 = check_shape(dist_params_q['logvar'], name='logvar_q', flatten=True)

            n = mu1.shape[-1]
            assert n == mu2.shape[-1] == logvar1.shape[-1] == logvar2.shape[-1]

            var1 = jnp.exp(logvar1)
            var2_inv = jnp.exp(-logvar2)
            logdetvar2 = jnp.sum(logvar2, axis=-1)  # log(det(M)) = tr(log(M))
            quadratic = jnp.einsum('ij,ij->i', var1 + jnp.square(mu1 - mu2), var2_inv)
            return 0.5 * (n * log_2pi + logdetvar2 + quadratic)

        def kl_divergence(dist_params_p, dist_params_q):
            mu1 = check_shape(dist_params_p['mu'], name='mu_p', flatten=True)
            mu2 = check_shape(dist_params_q['mu'], name='mu_q', flatten=True)
            logvar1 = check_shape(dist_params_p['logvar'], name='logvar_p', flatten=True)
            logvar2 = check_shape(dist_params_q['logvar'], name='logvar_q', flatten=True)

            n = mu1.shape[-1]
            assert n == mu2.shape[-1] == logvar1.shape[-1] == logvar2.shape[-1]

            var1 = jnp.exp(logvar1)
            var2_inv = jnp.exp(-logvar2)
            logdetvar1 = jnp.sum(logvar1, axis=-1)  # log(det(M)) = tr(log(M))
            logdetvar2 = jnp.sum(logvar2, axis=-1)  # log(det(M)) = tr(log(M))
            quadratic = jnp.einsum('ij,ij->i', var1 + jnp.square(mu1 - mu2), var2_inv)
            return 0.5 * (logdetvar2 - logdetvar1 + quadratic - n)

        def affine_transform_func(dist_params, scale, shift, value_transform=None):
            if value_transform is None:
                f = f_inv = lambda x: x
            else:
                f, f_inv = value_transform
            mu = check_shape(dist_params['mu'], name='mu', flatten=False)
            logvar = check_shape(dist_params['logvar'], name='logvar', flatten=False)
            var_new = f(f_inv(jnp.exp(logvar)) * jnp.square(scale))
            return {'mu': f(f_inv(mu) + shift), 'logvar': jnp.log(var_new)}

        self._sample_func = jit(sample)
        self._mean_func = jit(mean)
        self._mode_func = jit(mode)
        self._log_proba_func = jit(log_proba)
        self._entropy_func = jit(entropy)
        self._cross_entropy_func = jit(cross_entropy)
        self._kl_divergence_func = jit(kl_divergence)
        self._affine_transform_func = jit(affine_transform_func, static_argnums=(3,))

    @property
    def default_priors(self):
        shape = (1, *self.space.shape)  # include batch axis
        return {'mu': jnp.zeros(shape), 'logvar': jnp.zeros(shape)}

    def preprocess_variate(self, rng, X):
        X = jnp.asarray(X, dtype=self.space.dtype)                     # ensure ndarray
        X = jnp.reshape(X, (-1, *self.space.shape))                    # ensure batch axis
        X = jnp.clip(X, self._low, self._high)                         # clip to be safe
        X = clipped_logit((X - self._low) / (self._high - self._low))  # closed intervals -> reals
        return X

    def postprocess_variate(self, rng, X, index=0, batch_mode=False):
        X = jnp.asarray(X, dtype=self.space.dtype)                    # ensure ndarray
        X = jnp.reshape(X, (-1, *self.space.shape))                   # ensure correct shape
        X = jnp.clip(X, *self.clip_reals)                             # clip for stability
        X = self._low + (self._high - self._low) * jax.nn.sigmoid(X)  # reals -> closed interval
        return X if batch_mode else onp.asanyarray(X[index])

    @property
    def sample(self):
        r"""

        JIT-compiled function that generates differentiable variates using the reparametrization
        trick, i.e. :math:`x\sim\mathcal{N}(\mu,\sigma^2)` is implemented as

        .. math::

            \varepsilon\ &\sim\ \mathcal{N}(0,1) \\
            x\ &=\ \mu + \sigma\,\varepsilon

        Parameters
        ----------
        dist_params : pytree with ndarray leaves

            A batch of distribution parameters.

        rng : PRNGKey

            A key for seeding the pseudo-random number generator.

        Returns
        -------
        X : ndarray

            A batch of differentiable variates.

        """
        return self._sample_func

    @property
    def mean(self):
        r"""

        JIT-compiled functions that generates differentiable means of the distribution, in this case
        simply :math:`\mu`.

        Parameters
        ----------
        dist_params : pytree with ndarray leaves

            A batch of distribution parameters.

        Returns
        -------
        X : ndarray

            A batch of differentiable variates.

        """
        return self._mean_func

    @property
    def mode(self):
        r"""

        JIT-compiled functions that generates differentiable modes of the distribution, which for a
        normal distribution is the same as the :attr:`mean`.

        Parameters
        ----------
        dist_params : pytree with ndarray leaves

            A batch of distribution parameters.

        Returns
        -------
        X : ndarray

            A batch of differentiable variates.

        """
        return self._mode_func

    @property
    def log_proba(self):
        r"""

        JIT-compiled function that evaluates log-probabilities.

        Parameters
        ----------
        dist_params : pytree with ndarray leaves

            A batch of distribution parameters.

        X : ndarray

            A batch of variates, e.g. a batch of actions :math:`a` collected from experience.

        Returns
        -------
        logP : ndarray of floats

            A batch of log-probabilities associated with the provided variates.

        """
        return self._log_proba_func

    @property
    def entropy(self):
        r"""

        JIT-compiled function that computes the entropy of the distribution.

        .. math::

            H\ =\ -\mathbb{E}_p \log p
            \ =\ \frac12\left( \log(2\pi\sigma^2) + 1\right)

        Parameters
        ----------
        dist_params : pytree with ndarray leaves

            A batch of distribution parameters.

        Returns
        -------
        H : ndarray of floats

            A batch of entropy values.

        """
        return self._entropy_func

    @property
    def cross_entropy(self):
        r"""

        JIT-compiled function that computes the cross-entropy of a distribution :math:`q` relative
        to another categorical distribution :math:`p`:

        .. math::

            \text{CE}[p,q]\ =\ -\mathbb{E}_p \log q
            \ =\ \frac12\left(
                \log(2\pi\sigma_q^2)
                + \frac{(\mu_p-\mu_q)^2+\sigma_p^2}{\sigma_q^2}
            \right)

        Parameters
        ----------
        dist_params_p : pytree with ndarray leaves

            The distribution parameters of the *base* distribution :math:`p`.

        dist_params_q : pytree with ndarray leaves

            The distribution parameters of the *auxiliary* distribution :math:`q`.

        """
        return self._cross_entropy_func

    @property
    def kl_divergence(self):
        r"""

        JIT-compiled function that computes the Kullback-Leibler divergence of a categorical
        distribution :math:`q` relative to another distribution :math:`p`:

        .. math::

            \text{KL}[p,q]\ = -\mathbb{E}_p \left(\log q -\log p\right)
            \ =\ \frac12\left(
                \log(\sigma_q^2) - \log(\sigma_p^2)
                + \frac{(\mu_p-\mu_q)^2+\sigma_p^2}{\sigma_q^2}
                - 1
            \right)

        Parameters
        ----------
        dist_params_p : pytree with ndarray leaves

            The distribution parameters of the *base* distribution :math:`p`.

        dist_params_q : pytree with ndarray leaves

            The distribution parameters of the *auxiliary* distribution :math:`q`.

        """
        return self._kl_divergence_func



import gym
import jax.numpy as jnp
import numpy as onp

from ..utils import docstring
from .policy import Policy


__all__ = (
    'RandomPolicy',
)


class RandomPolicy:
    r"""

    A simple random policy.

    Parameters
    ----------
    env : gym.Env

        The gym-style environment. This is only used to get the :code:`env.action_space`.

    random_seed : int, optional

        Sets the random state to get reproducible results.

    """
    def __init__(self, env, random_seed=None):
        if not isinstance(env.action_space, gym.Space):
            raise TypeError(f"env.action_space must be a gym.Space, got: {type(env.action_space)}")
        self.action_space = env.action_space
        self.action_space.seed(random_seed)
        self.random_seed = random_seed

    @docstring(Policy.__call__)
    def __call__(self, s, return_logp=False):
        a = self.action_space.sample()
        if not return_logp:
            return a

        if isinstance(self.action_space, gym.spaces.Discrete):
            logp = -onp.log(self.num_actions)
            return a, logp

        if isinstance(self.action_space, gym.spaces.Box):
            sizes = self.action_space.high - self.action_space.low
            logp = -onp.sum(onp.log(sizes))  # log(prod(1/sizes))
            return a, logp

        raise NotImplementedError(
            "the log-propensity of a 'uniform' distribution over a "
            f"{self.action_space.__class__.__name__} space is not yet implemented; "
            "please submit a feature request")

    @docstring(Policy.mode)
    def mode(self, s):
        return self(s, return_logp=False)

    @docstring(Policy.dist_params)
    def dist_params(self, s):
        if isinstance(self.action_space, gym.spaces.Discrete):
            return {'logits': jnp.zeros(self.action_space.n)}

        if isinstance(self.action_space, gym.spaces.Box):
            return {
                'mu': jnp.zeros(self.action_space.shape),
                'logvar': 15 * jnp.ones(self.action_space.shape)}

        raise NotImplementedError(
            "the dist_params of a 'uniform' distribution over a "
            f"{self.action_space.__class__.__name__} space is not yet implemented; "
            "please submit a feature request")



import haiku as hk
import jax
import jax.numpy as jnp

from .._core.base_stochastic_func_type2 import BaseStochasticFuncType2
from ..utils import jit
from ._entropy import EntropyRegularizer


class NStepEntropyRegularizer(EntropyRegularizer):
    r"""

    Policy regularization term based on the n-step entropy of the policy.

    The regularization term is to be added to the loss function:

    .. math::

        \text{loss}(\theta; s,a)\ =\ -J(\theta; s,a) - \beta\,H[\pi_\theta(.|s)]

    where :math:`J(\theta)` is the bare policy objective.

    Parameters
    ----------
    f : stochastic function approximator

        The stochastic function approximator (e.g. :class:`coax.Policy`) to regularize.

    n : tuple(int), list(int), ndarray

        Time indices of the steps (counted from the current state at time `t`)
        to include in the regularization. For example `n = [2, 3]` adds an entropy bonus for the
        policy at the states t + 2 and t + 3 to the objective.

    beta : non-negative float

        The coefficient that determines the strength of the overall regularization term.

    gamma : float between 0 and 1

        The amount by which to discount the entropy bonuses.

    """

    def __init__(self, f, n, beta=0.001, gamma=0.99):
        super().__init__(f)
        if not isinstance(n, (tuple, list, jnp.ndarray)):
            raise TypeError(f"n must be a list, an ndarray or a tuple, got: {type(n)}")
        if len(n) == 0:
            raise ValueError("n cannot be empty")
        self.n = n
        self._n = jnp.array(n)
        self.beta = beta
        self.gamma = gamma
        self._gammas = jnp.take(jnp.power(self.gamma, jnp.arange(self.n[-1] + 1)), self._n)

        def entropy(dist_params, dones):
            valid = self.valid_from_done(dones)
            vf = jax.vmap(lambda p, v, gamma: gamma * self.f.proba_dist.entropy(p) * v)
            return jnp.sum(vf(dist_params, valid, self._gammas), axis=0)

        def function(dist_params, dones, beta):
            assert len(dist_params) == 2
            return -beta * entropy(dist_params, dones)

        def metrics(dist_params, dones, beta):
            assert len(dist_params) == 2
            valid = self.valid_from_done(dones)
            return {
                'EntropyRegularizer/beta': beta,
                'EntropyRegularizer/entropy': jnp.mean(entropy(dist_params, dones) /
                                                       jnp.sum(valid, axis=0))
            }

        self._function = jit(function)
        self._metrics_func = jit(metrics)

    @property
    def batch_eval(self):
        if not hasattr(self, '_batch_eval_func'):
            def batch_eval_func(params, hyperparams, state, rng, transition_batch):
                rngs = hk.PRNGSequence(rng)
                if not isinstance(transition_batch.extra_info, dict):
                    raise TypeError(
                        'TransitionBatch.extra_info has to be a dict containing "states" and' +
                        ' "dones" for the n-step entropy regularization. Make sure to set the' +
                        ' record_extra_info flag in the NStep tracer.')
                if isinstance(self.f, BaseStochasticFuncType2):
                    def f(s_next):
                        return self.f.function(params,
                                               state, next(rngs),
                                               self.f.observation_preprocessor(
                                                   next(rngs), s_next), True)
                    n_states = transition_batch.extra_info['states']
                    dist_params, _ = jax.vmap(f)(jax.tree_util.tree_map(
                        lambda *t: jnp.stack(t), *n_states))
                    dist_params = jax.tree_util.tree_map(
                        lambda t: jnp.take(t, self._n, axis=0), dist_params)
                else:
                    raise TypeError(
                        "f must be derived from BaseStochasticFuncType2")
                dones = jnp.stack(transition_batch.extra_info['dones'])
                dones = jnp.take(dones, self._n, axis=0)

                return self.function(dist_params,
                                     dones,
                                     **hyperparams), self.metrics_func(dist_params, dones,
                                                                       **hyperparams)

            self._batch_eval_func = jit(batch_eval_func)

        return self._batch_eval_func

    def valid_from_done(self, dones):
        """
        Generates a mask that filters all time steps after a done signal has been reached.

        Parameters
        ----------
        dones : ndarray

            Array of boolean entries indicating whether the episode has ended.

        Returns
        -------
        valid : ndarray

            Mask that filters all entries after a done=True has been reached.
        """
        valid = jnp.ones_like(dones, dtype=jnp.float32)
        return valid.at[1:].set(1 - jnp.clip(jnp.cumsum(dones[:-1], axis=0), a_max=1))




from inspect import signature

import jax


__all__ = (
    'JittedFunc',
    'jit',
)


def jit(func, static_argnums=(), donate_argnums=()):
    r"""

    An alternative of :func:`jax.jit` that returns a picklable JIT-compiled function.

    Note that :func:`jax.jit` produces non-picklable functions, because the JIT compilation depends
    on the :code:`device` and :code:`backend`. In order to facilitate serialization, this function
    does not for the user to specify :code:`device` or :code:`backend`. Instead :func:`jax.jit` is
    called with the default: :code:`jax.jit(..., device=None, backend=None)`.

    Check out the original :func:`jax.jit` docs for a more detailed description of the arguments.

    Parameters
    ----------
    func : function

        Function to be JIT compiled.

    static_argnums : int or tuple of ints

        Arguments to exclude from JIT compilation.

    donate_argnums : int or tuple of ints

        To be donated arguments, see :func:`jax.jit`.

    Returns
    -------
    jitted_func : JittedFunc

        A picklable JIT-compiled function.

    """
    return JittedFunc(func, static_argnums, donate_argnums)


class JittedFunc:
    __slots__ = ('func', 'static_argnums', 'donate_argnums', '_jitted_func')

    def __init__(self, func, static_argnums=(), donate_argnums=()):
        self.func = func
        self.static_argnums = static_argnums
        self.donate_argnums = donate_argnums
        self._init_jitted_func()

    def __call__(self, *args, **kwargs):
        return self._jitted_func(*args, **kwargs)

    @property
    def __signature__(self):
        return signature(self.func)

    def __repr__(self):
        return self.__class__.__name__ + str(self.__signature__)

    def __getstate__(self):
        return self.func, self.static_argnums, self.donate_argnums

    def __setstate__(self, state):
        self.func, self.static_argnums, self.donate_argnums = state
        self._init_jitted_func()

    def _init_jitted_func(self):
        self._jitted_func = jax.jit(
            self.func,
            static_argnums=self.static_argnums,
            donate_argnums=self.donate_argnums)



import haiku as hk
from jax.scipy.special import logsumexp
from gym.spaces import Discrete

from ..utils import is_stochastic
from ._base import BaseTDLearningQ


class SoftQLearning(BaseTDLearningQ):
    r"""

    TD-learning with soft Q-learning updates. The :math:`n`-step bootstrapped target is constructed
    as:

    .. math::

        G^{(n)}_t\ =\ R^{(n)}_t
            + I^{(n)}_t\,\tau\log\sum_{a'}\exp\left(q_\text{targ}(S_{t+n}, a') / \tau\right)

    where

    .. math::

        R^{(n)}_t\ &=\ \sum_{k=0}^{n-1}\gamma^kR_{t+k} \\
        I^{(n)}_t\ &=\ \left\{\begin{matrix}
            0           & \text{if $S_{t+n}$ is a terminal state} \\
            \gamma^n    & \text{otherwise}
        \end{matrix}\right.

    Parameters
    ----------
    q : Q

        The main q-function to update.

    q_targ : Q, optional

        The q-function that is used for constructing the TD-target. If this is left unspecified, we
        set ``q_targ = q`` internally.

    optimizer : optax optimizer, optional

        An optax-style optimizer. The default optimizer is :func:`optax.adam(1e-3)
        <optax.adam>`.

    loss_function : callable, optional

        The loss function that will be used to regress to the (bootstrapped) target. The loss
        function is expected to be of the form:

        .. math::

            L(y_\text{true}, y_\text{pred}, w)\in\mathbb{R}

        where :math:`w>0` are sample weights. If left unspecified, this defaults to
        :func:`coax.value_losses.huber`. Check out the :mod:`coax.value_losses` module for other
        predefined loss functions.

    policy_regularizer : Regularizer, optional

        If provided, this policy regularizer is added to the TD-target. A typical example is to use
        an :class:`coax.regularizers.EntropyRegularizer`, which adds the policy entropy to
        the target. In this case, we minimize the following loss shifted by the entropy term:

        .. math::

            L(y_\text{true} + \beta\,H[\pi], y_\text{pred})

        Note that the coefficient :math:`\beta` plays the role of the temperature in SAC-style
        agents.

    temperature : float, optional

        The Boltzmann temperature :math:`\tau>0`.

    """
    def __init__(
            self, q, q_targ=None, optimizer=None, loss_function=None, policy_regularizer=None,
            temperature=1.0):

        if not isinstance(q.action_space, Discrete):
            raise NotImplementedError(
                f"{self.__class__.__name__} class is only implemented for discrete actions spaces")
        if is_stochastic(q):
            raise NotImplementedError(f"{type(self).__name__} is not yet implement for StochasticQ")

        self.temperature = temperature
        super().__init__(
            q=q,
            q_targ=q_targ,
            optimizer=optimizer,
            loss_function=loss_function,
            policy_regularizer=policy_regularizer)

    def target_func(self, target_params, target_state, rng, transition_batch):
        rngs = hk.PRNGSequence(rng)
        params, state = target_params['q_targ'], target_state['q_targ']
        S_next = self.q_targ.observation_preprocessor(next(rngs), transition_batch.S_next)
        Q_s_next, _ = self.q_targ.function_type2(params, state, next(rngs), S_next, False)
        assert Q_s_next.ndim == 2
        assert Q_s_next.shape[1] == self.q.action_space.n
        Q_sa_next = self.temperature * logsumexp(Q_s_next / self.temperature, axis=-1)
        assert Q_sa_next.ndim == 1

        f, f_inv = self.q.value_transform.transform_func, self.q_targ.value_transform.inverse_func
        return f(transition_batch.Rn + transition_batch.In * f_inv(Q_sa_next))



import jax.numpy as jnp

from ._base import ValueTransform


class LogTransform(ValueTransform):
    r"""

    A simple invertible log-transform.


    .. math::

        x\ \mapsto\ y\ =\ \lambda\,\text{sign}(x)\,
            \log\left(1+\frac{|x|}{\lambda}\right)

    with inverse:

    .. math::

        y\ \mapsto\ x\ =\ \lambda\,\text{sign}(y)\,
            \left(\text{e}^{|y|/\lambda} - 1\right)

    This transform logarithmically supresses large values :math:`|x|\gg1` and smoothly interpolates
    to the identity transform for small values :math:`|x|\sim1` (see figure below).

    .. image:: /_static/img/log_transform.svg
        :alt: Invertible log-transform
        :width: 640px

    Parameters
    ----------
    scale : positive float, optional

        The scale :math:`\lambda>0` of the linear-to-log cross-over. Smaller
        values for :math:`\lambda` translate into earlier onset of the
        cross-over.

    """
    __slots__ = ValueTransform.__slots__ + ('scale',)

    def __init__(self, scale=1.0):
        assert scale > 0
        self.scale = scale

        def transform_func(x):
            return jnp.sign(x) * scale * jnp.log(1 + jnp.abs(x) / scale)

        def inverse_func(x):
            return jnp.sign(x) * scale * (jnp.exp(jnp.abs(x) / scale) - 1)

        self._transform_func = transform_func
        self._inverse_func = inverse_func



import jax
import jax.numpy as jnp


__all__ = (
    'mse',
    'huber',
    'logloss',
    'logloss_sign',
)


def mse(y_true, y_pred, w=None):
    r"""

    Ordinary mean-squared error loss function.

    .. math::

        L\ =\ \frac12(\hat{y} - y)^2

    .. image:: /_static/img/mse.svg
        :alt: Mean-Squared Error loss
        :width: 320px
        :align: center

    Parameters
    ----------
    y_true : ndarray

        The target :math:`y\in\mathbb{R}`.

    y_pred : ndarray

        The predicted output :math:`\hat{y}\in\mathbb{R}`.

    w : ndarray, optional

        Sample weights.

    Returns
    -------
    loss : scalar ndarray

        The loss averaged over the batch.

    """
    loss = 0.5 * jnp.square(y_pred - y_true)
    return _mean_with_weights(loss, w)


def huber(y_true, y_pred, w=None, delta=1.0):
    r"""

    `Huber <https://en.wikipedia.org/wiki/Huber_loss>`_ loss function.

    .. math::

        L\ =\ \left\{\begin{matrix}
                (\hat{y} - y)^2
                    &\quad:\ |\hat{y} - y|\leq\delta \\
                \delta\,|\hat{y} - y| - \frac{\delta^2}{2}
                    &\quad:\ |\hat{y} - y| > \delta
            \end{matrix}\right.

    .. image:: /_static/img/huber.svg
        :alt: Huber loss
        :width: 320px
        :align: center

    Parameters
    ----------
    y_true : ndarray

        The target :math:`y\in\mathbb{R}`.

    y_pred : ndarray

        The predicted output :math:`\hat{y}\in\mathbb{R}`.

    w : ndarray, optional

        Sample weights.

    delta : float, optional

        The scale of the quadratic-to-linear transition.

    Returns
    -------
    loss : scalar ndarray

        The loss averaged over the batch.

    """
    err = jnp.abs(y_pred - y_true)
    err_clipped = jnp.minimum(err, delta)
    loss = 0.5 * jnp.square(err_clipped) + delta * (err - err_clipped)
    return _mean_with_weights(loss, w)


def logloss(y_true, y_pred, w=None):
    r"""

    Logistic loss function for binary classification, `y_true` =
    :math:`y\in\{0,1\}` and the model output is a probability `y_pred` =
    :math:`\hat{y}\in[0,1]`:

    .. math::

        L\ =\ -y\log(\hat{y}) - (1 - y)\log(1 - \hat{y})

    Parameters
    ----------
    y_true : ndarray

        The binary target, encoded as :math:`y\in\{0,1\}`.

    y_pred : (ndarray of) float

        The predicted output, represented by a probablity
        :math:`\hat{y}\in[0,1]`.

    w : ndarray, optional

        Sample weights.

    Returns
    -------
    loss : scalar ndarray

        The loss averaged over the batch.

    """
    loss = -y_true * jnp.log(y_pred) - (1. - y_true) * jnp.log(1. - y_pred)
    return _mean_with_weights(loss, w)


def logloss_sign(y_true_sign, logits, w=None):
    r"""

    Logistic loss function specific to the case in which the target is a sign
    :math:`y\in\{-1,1\}` and the model output is a logit
    :math:`\hat{z}\in\mathbb{R}`.

    .. math::

        L\ =\ \log(1 + \exp(-y\,\hat{z}))

    This version tends to be more numerically stable than the generic
    implementation, because it avoids having to map the predicted logit to a
    probability.

    Parameters
    ----------
    y_true_sign : ndarray

        The binary target, encoded as :math:`y=\pm1`.

    logits : ndarray

        The predicted output, represented by a logit
        :math:`\hat{z}\in\mathbb{R}`.

    w : ndarray, optional

        Sample weights.

    Returns
    -------
    loss : scalar ndarray

        The loss averaged over the batch.

    """
    loss = jnp.log(1.0 + jnp.exp(-y_true_sign * logits))
    return _mean_with_weights(loss, w)


def _mean_with_weights(loss, w):
    if w is not None:
        assert w.ndim == 1
        assert loss.ndim >= 1
        assert loss.shape[0] == w.shape[0]
        loss = jax.vmap(jnp.multiply)(w, loss)
    return jnp.mean(loss)


def quantile_huber(y_true, y_pred, quantiles, w=None, delta=1.0):
    r"""

    `Quantile Huber <https://arxiv.org/abs/1806.06923>`_ loss function.

    .. math::

        \delta_{ij} &= y_j - \hat{y}_i\\
        \rho^\kappa_\tau(\delta_{ij}) &= |\tau - \mathbb{I}{\{ \delta_{ij} < 0 \}}| \
            \frac{\mathcal{L}_\kappa(\delta_{ij})}{\kappa},\ \quad \text{with}\\
        \mathcal{L}_\kappa(\delta_{ij}) &= \begin{cases}
            \frac{1}{2} \delta_{ij}^2,\quad \ &\text{if } |\delta_{ij}| \le \kappa\\
            \kappa (|\delta_{ij}| - \frac{1}{2}\kappa),\quad \ &\text{otherwise}
        \end{cases}

    Parameters
    ----------
    y_true : ndarray

        The target :math:`y\in\mathbb{R}^{2}`.

    y_pred : ndarray

        The predicted output :math:`\hat{y}\in\mathbb{R}^{2}`.

    quantiles : ndarray

        The quantiles of the prediction :math:`\tau\in\mathbb{R}^{2}`.

    w : ndarray, optional

        Sample weights.

    delta : float, optional

        The scale of the quadratic-to-linear transition.

    Returns
    -------
    loss : scalar ndarray

        The loss averaged over the batch.

    """
    y_pred = y_pred[..., None]
    y_true = y_true[..., None, :]
    quantiles = quantiles[..., None]
    td_error = y_true - y_pred
    td_error_abs = jnp.abs(td_error)
    err_clipped = jnp.minimum(td_error_abs, delta)
    elementwise_huber_loss = 0.5 * jnp.square(err_clipped) + delta * (td_error_abs - err_clipped)
    elementwise_quantile_huber_loss = jnp.abs(
        quantiles - (td_error < 0)) * elementwise_huber_loss / delta
    quantile_huber_loss = elementwise_quantile_huber_loss.sum(axis=-1)
    return _mean_with_weights(quantile_huber_loss, w=w)



import haiku as hk

from ..utils import is_stochastic
from ._base import BaseTDLearningQ


class Sarsa(BaseTDLearningQ):
    r"""

    TD-learning with SARSA updates. The :math:`n`-step bootstrapped target is constructed as:

    .. math::

        G^{(n)}_t\ =\ R^{(n)}_t + I^{(n)}_t\,q_\text{targ}(S_{t+n}, A_{t+n})

    where :math:`A_{t+n}` is sampled from experience and

    .. math::

        R^{(n)}_t\ &=\ \sum_{k=0}^{n-1}\gamma^kR_{t+k} \\
        I^{(n)}_t\ &=\ \left\{\begin{matrix}
            0           & \text{if $S_{t+n}$ is a terminal state} \\
            \gamma^n    & \text{otherwise}
        \end{matrix}\right.

    Parameters
    ----------
    q : Q

        The main q-function to update.

    q_targ : Q, optional

        The q-function that is used for constructing the TD-target. If this is left unspecified, we
        set ``q_targ = q`` internally.

    optimizer : optax optimizer, optional

        An optax-style optimizer. The default optimizer is :func:`optax.adam(1e-3)
        <optax.adam>`.

    loss_function : callable, optional

        The loss function that will be used to regress to the (bootstrapped) target. The loss
        function is expected to be of the form:

        .. math::

            L(y_\text{true}, y_\text{pred}, w)\in\mathbb{R}

        where :math:`w>0` are sample weights. If left unspecified, this defaults to
        :func:`coax.value_losses.huber`. Check out the :mod:`coax.value_losses` module for other
        predefined loss functions.

    policy_regularizer : Regularizer, optional

        If provided, this policy regularizer is added to the TD-target. A typical example is to use
        an :class:`coax.regularizers.EntropyRegularizer`, which adds the policy entropy to
        the target. In this case, we minimize the following loss shifted by the entropy term:

        .. math::

            L(y_\text{true} + \beta\,H[\pi], y_\text{pred})

        Note that the coefficient :math:`\beta` plays the role of the temperature in SAC-style
        agents.

    """
    def target_func(self, target_params, target_state, rng, transition_batch):
        rngs = hk.PRNGSequence(rng)
        params, state = target_params['q_targ'], target_state['q_targ']
        S_next = self.q_targ.observation_preprocessor(next(rngs), transition_batch.S_next)
        A_next = self.q_targ.action_preprocessor(next(rngs), transition_batch.A_next)

        if is_stochastic(self.q):
            return self._get_target_dist_params(params, state, next(rngs), transition_batch, A_next)

        Q_sa_next, _ = self.q_targ.function_type1(params, state, next(rngs), S_next, A_next, False)
        f, f_inv = self.q.value_transform.transform_func, self.q_targ.value_transform.inverse_func
        return f(transition_batch.Rn + transition_batch.In * f_inv(Q_sa_next))
